from typing import Dict, Optional, Union

from pyspark.sql import DataFrame
from pyspark.sql.streaming import DataStreamWriter

from karadoc.common.conf import CONNECTION_GROUP


def _write_stream_external_output_signature_check(df: DataFrame, dest: Dict) -> DataStreamWriter:
    """We uses this empty method when to check the signature of the equivalent method defined in the STREAM files"""
    pass


class HasStreamExternalOutput:
    def __init__(self) -> None:
        # Attributes that the user may change
        self.external_output: Optional[dict] = None

        # Attributes that the user is not supposed to change
        self.__write_external_output = None

    def _write_external_output_default(self, df: DataFrame, dest: Dict) -> DataStreamWriter:
        connector = self.get_output_connector(dest)
        return connector.write_stream(df, dest)

    def write_external_output(self, df: DataFrame, dest: Dict) -> DataStreamWriter:
        """Writes a given DataFrame to a given external output

        :param df: The DataFrame to write
        :param dest: The alias of the external output to write to
        :return: nothing
        """
        if self.__write_external_output is None:
            return self._write_external_output_default(df, dest)
        else:
            return self.__write_external_output(df, dest)

    def get_output_connector(self, dest: Union[str, Dict]):
        if type(dest) == str:
            dest = self.external_output[dest]
        from karadoc.common.connector import load_connector

        return load_connector(dest[CONNECTION_GROUP], self.spark)



from typing import Dict, Optional

from pyspark.sql import DataFrame

from karadoc.common.job_core.load import (
    load_non_runnable_action_file,
    load_runnable_action_file,
)
from karadoc.common.output.printing import pretty_format
from karadoc.common.run.spark_batch_job import SparkBatchJob
from karadoc.common.utils.assert_utils import assert_true


def load_populate(full_table_name: str) -> SparkBatchJob:
    """Return a non-runnable version of the `job` object declared in the POPULATE.py file of the given table
    with its default variables.

    This method is useful to load any POPULATE for inspecting its metadata (input, output, output_partitions, etc.)
    even when no variable is defined, by using its default values. The `run` method is removed to prevent any misuse
    where one would call the POPULATE without setting the variables properly.

    If you want to execute the `run` method use the `load_runnable_populate` method instead.

    :param full_table_name: Full name of the table (schema_name.table_name)
    :return: the SparkBatchJob object defined in the POPULATE.py file
    """
    return load_non_runnable_action_file(full_table_name, SparkBatchJob)


def load_runnable_populate(full_table_name: str, passed_vars: Optional[Dict[str, str]]) -> SparkBatchJob:
    """Return the `job` object declared in the POPULATE.py file of the given table, override its variables
    and check the load variables match those declared in the POPULATE.py file.

    :param full_table_name: Full name of the table (schema_name.table_name)
    :param passed_vars: Set the variables of the job and check that they match the ones declared
    :return: the SparkBatchJob object defined in the POPULATE.py file
    """
    return load_runnable_action_file(full_table_name, SparkBatchJob, passed_vars)


def print_job_properties(job: SparkBatchJob):
    if job.inputs:
        print(f"inputs : {job.inputs}")
        print(f"input_paths : {job.list_hdfs_inputs()}")
    if job.external_inputs:
        print("external_inputs :")
        print(pretty_format(job.external_inputs))
    if job.output:
        print(f"output : {job.output}")
        if job.output_partition:
            print(f"output_partition : {job.output_partition}")
        print(f"output_path : {job.hdfs_output()}")
    if job.external_outputs:
        print("external_outputs:")
        print(pretty_format(job.external_outputs))


def write_output(job: SparkBatchJob, df: DataFrame, no_write: bool = False, no_export: bool = False):
    """Serialize the output of a SparkBatchJob if it exists and is not empty, and if its schema is correctly defined.

    :param no_write:
    """
    if df is None:
        return

    assert_true(isinstance(df, DataFrame))

    if len(df.columns) > 0:
        if no_write:
            df.count()  # We force the DataFrame evaluation
            print("WARN: outputs have been disabled.")
        if len(job.external_outputs) > 0:
            if no_export:
                print("WARN: exports have been disabled.")
            else:
                job.write_external_outputs(df)
        else:
            job.write_table(df)
    else:
        print("WARN: The job returned a DataFrame with zero columns. The output will not be serialized.")



import ast
import re
from datetime import datetime, timedelta
from typing import Dict, List

from karadoc.common.utils.assert_utils import assert_true

DATE_FORMAT = "%Y-%m-%d"


def day_range(start_date: str, end_date: str, step: int = 1) -> List[str]:
    """Compute the interval of days [start_date, end_date[
    >>> day_range('2019-10-01', '2019-10-07', 1)
    ['2019-10-01', '2019-10-02', '2019-10-03', '2019-10-04', '2019-10-05', '2019-10-06']
    >>> day_range('2019-10-01', '2019-10-31', 7)
    ['2019-10-01', '2019-10-08', '2019-10-15', '2019-10-22', '2019-10-29']

    :param start_date: Left bound of the interval (included)
    :param end_date: Right bound of the interval (not included). Should be greater or equal to start_date
    :param step: (default: 1) Number of days between each element
    :return:
    """
    assert_true(step > 0, "day_range() arg step should be greater than zero")
    start_datetime = datetime.strptime(start_date, DATE_FORMAT)
    end_datetime = datetime.strptime(end_date, DATE_FORMAT)
    numdays = (end_datetime - start_datetime).days
    assert_true(start_datetime <= end_datetime, "day_range() arg end_date should be greater or equal to start_date")
    return [(start_datetime + timedelta(days=x)).strftime(DATE_FORMAT) for x in range(0, numdays, step)]


def expand_vars(vars: Dict[str, str]) -> List[Dict[str, str]]:
    """

    >>> vars = { "day_1": "day_range('2019-10-01', '2019-10-03')", "day_2": "day_range('2020-01-01', '2020-01-03')"}
    >>> for v in expand_vars(vars):
    ...     print(v)
    {'day_1': '2019-10-01', 'day_2': '2020-01-01'}
    {'day_1': '2019-10-01', 'day_2': '2020-01-02'}
    {'day_1': '2019-10-02', 'day_2': '2020-01-01'}
    {'day_1': '2019-10-02', 'day_2': '2020-01-02'}

    :param vars:
    :return:
    """
    res = [vars.copy()]
    for key, value in vars.items():
        values = [value]
        match = next(re.finditer(r"^day_range(?P<args>\(.*\))$", value), None)
        if match is not None:
            args = match.group("args")
            values = day_range(*ast.literal_eval(args))
        new_res = []
        for r in res:
            for v in values:
                new_r = r.copy()
                new_r[key] = v
                new_res.append(new_r)
        res = new_res
    return res



import glob
import importlib
import logging as _logging
import pkgutil
import re
import sys
import traceback
from argparse import ArgumentParser, Namespace
from os.path import basename, dirname, isfile
from typing import Dict, List, Optional, Sequence, Type, Union

from karadoc.common import conf
from karadoc.common.class_utils import find_class_from_module
from karadoc.common.commands.command import Command
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.conf import APPLICATION_DESCRIPTION, APPLICATION_NAME
from karadoc.common.exceptions import CommandValidationError, LoggedException
from karadoc.common.observability import LogEvent, get_logging_context
from karadoc.common.observability.conf import setup_observability
from karadoc.common.observability.console_event import ConsoleEvent
from karadoc.common.utils.assert_utils import assert_true
from karadoc.common.utils.stopwatch import Stopwatch

COMMAND_ROOT_DEPTH = 0
BUILTIN_COMMAND_PACKAGE = "karadoc.cli.commands"
modules = glob.glob(dirname(__file__) + "/*.py")
__all__ = [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py")]
LOG = _logging.getLogger(__name__)


def __full_name(command: Type[Command]) -> str:
    return f"{command.__module__}.{command.__name__}"


def _check_not_abstract(command: Type[Command]):
    """Raise an exception if the class is abstract"""
    abstract_methods = set(getattr(command, "__abstractmethods__"))
    if len(abstract_methods) > 0:
        raise CommandValidationError(
            f"The class {__full_name(command)} should implement the following methods: {abstract_methods}"
        )


def _print_command_override_warning(
    command_name: str, overridden_command_full_name: str, overriding_command_full_name: str
) -> None:
    print(
        f"WARNING: command '{command_name}', "
        f"initially defined as {overridden_command_full_name}, "
        f"has been replaced with {overriding_command_full_name}"
    )


def _merge_command_packages(command_packages_with_names: List[Dict[str, Type[Command]]]) -> Dict[str, Type[Command]]:
    """Merge together several command packages, and print warnings whenever a command would be
    overridden by another one."""
    assert_true(len(command_packages_with_names) > 0, "No command package found")
    result = command_packages_with_names[0]
    for command_package in command_packages_with_names[1:]:
        for command_name, command in command_package.items():
            if command_name in result:
                _print_command_override_warning(
                    command_name=command_name,
                    overridden_command_full_name=__full_name(result[command_name]),
                    overriding_command_full_name=__full_name(command),
                )
            result[command_name] = command
    return result


def load_commands() -> Dict[str, Type[Command]]:
    """Load all commands from built-in and custom packages"""
    packages = [BUILTIN_COMMAND_PACKAGE] + conf.get_custom_command_packages()
    command_packages = [load_commands_from_package(package, command_depth=COMMAND_ROOT_DEPTH) for package in packages]
    merged_command_package = _merge_command_packages(command_packages)
    return merged_command_package


def load_commands_from_package(package_name: str, command_depth: int) -> Dict[str, Type[Command]]:
    """Automagically load all commands that are in specified module

    :param package_name: path to the python package or module containing the command
    :param command_depth: depth of the command
      (e.g. in "karadoc build package", 'build' has depth 0 and 'package' has depth 1)
    :return: a Dict(command_name, command)
    """

    def _load_commands():
        commands = importlib.import_module(package_name)
        for importer, modname, ispkg in pkgutil.iter_modules(commands.__path__):
            module_full_path = f"{package_name}.{modname}"
            module = importlib.import_module(module_full_path)
            command_class = find_class_from_module(module, Command)
            command_class.command_depth = command_depth
            yield command_class

    return {command._get_command_name(): command for command in _load_commands()}


def _dest(command_depth: int) -> str:
    """Return the name of the "args" attribute in which the command or subcommand will be stored, depending on its
    depth level.

    Explanation
    -----------

    As shown in the example below, the 'dest' attribute is set to command_0 for the command group and command_1 for
    the subcommand. They have to be distinct because they are both added to the flat `args` object returned by
    `parser.parse_args`. We need to use the `command_depth` to avoid collision between commands, subcommands,
    subsubcommands, etc.

    >>> parser = ArgumentParser()
    >>> sub_parsers = parser.add_subparsers(title='commands', dest='command_0')
    >>> sub_parser = sub_parsers.add_parser("custom_command_group")
    >>> sub_sub_parsers = sub_parser.add_subparsers(title='commands', dest='command_1')
    >>> sub_sub_parser = sub_sub_parsers.add_parser("sub_command")
    >>> args = parser.parse_args(["custom_command_group", "sub_command"])
    >>> args.command_0
    'custom_command_group'
    >>> args.command_1
    'sub_command'
    """
    return f"command_{command_depth}"


def add_commands_to_parser(parser: ArgumentParser, commands: Dict[str, Type[Command]], command_depth: int) -> None:
    dest = _dest(command_depth)
    subparsers = parser.add_subparsers(title="commands", metavar="", prog=APPLICATION_NAME.lower(), dest=dest)
    for command_name, command in commands.items():
        _check_not_abstract(command)
        subparser = subparsers.add_parser(command_name, help=command.description)
        subparser.description = command.description
        command.add_arguments(subparser)


def _split_command_line(command_line: str) -> List[str]:
    """Split a command string by whitespace characters

    :param command_line:
    :return:
    """
    return re.compile("\\s+").split(command_line)


def do_command(
    args: Namespace, parser: ArgumentParser, commands_by_name: Dict[str, Type[Command]], command_depth: int
) -> Optional[ReturnCode]:
    command_to_run = getattr(args, _dest(command_depth))
    if command_to_run is not None:
        assert_true(
            (command_to_run in commands_by_name),
            f"Probable Bug: command {command_to_run} was found by arg_parser but is not listed in commands_by_name",
        )
        return commands_by_name[command_to_run].do_command(args)
    else:
        parser.print_help()


def __run_command(argv: List[str]) -> ReturnCode:
    parser = ArgumentParser(description=APPLICATION_DESCRIPTION, prog=APPLICATION_NAME.lower())
    parser.add_argument("-v", "--verbose", default=False, action="store_true", help="Activate verbose mode")
    commands_by_name = load_commands()
    add_commands_to_parser(parser, commands_by_name, command_depth=COMMAND_ROOT_DEPTH)
    args = parser.parse_args(argv)
    args.raw_args = " ".join(argv)

    return_value = do_command(args, parser, commands_by_name, command_depth=COMMAND_ROOT_DEPTH)
    if return_value is not None:
        return return_value
    else:
        return ReturnCode.Success


def _get_command_end_extra_info(return_code: ReturnCode, stopwatch: Stopwatch) -> Dict[str, str]:
    return {"return_code": str(return_code), "duration_in_seconds": stopwatch.duration_in_seconds}


def _run_command_unsafe(command_line: Union[str, Sequence[str]]) -> ReturnCode:
    return_code = ReturnCode.Error
    stopwatch = Stopwatch()
    argv = _split_command_line(command_line) if isinstance(command_line, str) else command_line
    command = " ".join(argv)
    # Workaround: we detect the --verbose option as soon as possible before starting to log events.
    if "-v" in argv or "--verbose" in argv:
        verbose = True
    else:
        verbose = False
    setup_observability(reset=True, verbose=verbose)
    get_logging_context().command = command
    try:
        LOG.info(LogEvent(f"Command started: {command}", event_type="command_start"))
        return_code = __run_command(argv)
        extra = _get_command_end_extra_info(return_code, stopwatch)
        LOG.info(LogEvent(f"Command ended: {command}", event_type="command_end", extra=extra))
    except LoggedException as err:
        # If the exception has already been logged, we don't send a ConsoleEvent.
        extra = _get_command_end_extra_info(return_code, stopwatch)
        LOG.error(LogEvent(f"Command failed: {command}", event_type="command_end", extra=extra), exc_info=err)
        raise err
    except Exception as err:
        extra = _get_command_end_extra_info(return_code, stopwatch)
        LOG.error(ConsoleEvent(f"Command failed: {command}", event_type="command_end", extra=extra), exc_info=err)
        raise LoggedException() from err
    return return_code


def run_command(command_line: Union[str, Sequence[str]]) -> ReturnCode:
    """Execute a karadoc command line

    Examples:

    - run_command("run --tables s.t1 s.t2")
    - run_command(["run", "--tables", "s.t1", "s.t2"])

    :param command_line: A plain string command, or a list of arguments
    """
    try:
        return _run_command_unsafe(command_line)
    except LoggedException as err:
        raise err.__cause__
    except Exception as err:
        print("ERROR: An exception occurred before logging could be configured", file=sys.stderr)
        traceback.print_exc()
        raise err



from argparse import ArgumentParser, Namespace

from karadoc.common import conf
from karadoc.common.analyze import draw_timeline_analysis, timeline_analysis
from karadoc.common.analyze.analyze_job import AnalyzeJob
from karadoc.common.commands.command import Command
from karadoc.common.commands.options.dry_option import DryOption
from karadoc.common.commands.options.read_from_option import ReadFromOption
from karadoc.common.commands.options.spark_conf_option import SparkConfOption
from karadoc.common.commands.options.tables_option import TablesOption
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.commands.spark import init_job
from karadoc.common.job_core.load import load_runnable_action_file
from karadoc.common.table_utils import get_data_location
from karadoc.common.utils.assert_utils import assert_true
from karadoc.common.validations import fail_if_results
from karadoc.common.validations.job_validations import validate_inputs


def _print_job_properties(job: AnalyzeJob):
    if job.inputs:
        print("inputs : %s" % job.inputs)


def _check_job(job: AnalyzeJob):
    assert_true(job.reference_time_col is not None, "reference_time_col is not defined")
    assert_true(job.cohorts is not None, "cohort is not defined")
    assert_true(job.nb_buckets is not None, "nb_buckets is not defined")
    assert_true(job.analyze is not None, "job.analyze is not defined")


class AnalyzeTimelineCommand(Command):
    description = "run timeline analysis scripts for the specified tables"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        TablesOption.add_arguments(parser)
        DryOption.add_arguments(parser)
        ReadFromOption.add_arguments(parser)
        SparkConfOption.add_arguments(parser)

    @staticmethod
    def do_command(args: Namespace) -> ReturnCode:
        for table in args.tables:
            print("starting analyzing table %s" % table)
            data_location = get_data_location(table)
            print(data_location)
            job = load_runnable_action_file(table, AnalyzeJob, {})
            _print_job_properties(job)
            fail_if_results(validate_inputs(job))
            _check_job(job)
            if not args.dry:
                init_job(job=job, raw_args=args.raw_args, spark_conf=args.spark_conf)
                if args.remote_env is not None:
                    job._configure_remote_input(args.remote_env)
                df = job.analyze()
                timeline = timeline_analysis.analyse_timeline(
                    df, reference_time_col=job.reference_time_col, cohort_cols=job.cohorts, nb_buckets=job.nb_buckets
                )
                output_dir = conf.get_analyze_timeline_dir_location()
                if len(job.cohorts) != 1:
                    draw_timeline_analysis.draw(timeline, job.cohorts, output_dir, table)
                else:
                    draw_timeline_analysis.draw(timeline, job.cohorts + [""], output_dir, table)

        return ReturnCode.Success



from karadoc.common.job_core.has_disable import HasDisable
from karadoc.common.job_core.has_spark import HasSpark
from karadoc.common.job_core.has_stream_external_inputs import HasStreamExternalInputs
from karadoc.common.job_core.has_stream_external_output import HasStreamExternalOutput
from karadoc.common.job_core.has_stream_inputs import HasStreamInputs
from karadoc.common.job_core.has_stream_output import HasStreamOutput
from karadoc.common.job_core.has_vars import HasVars


class SparkStreamJob(
    HasVars,
    HasStreamOutput,
    HasStreamInputs,
    HasStreamExternalInputs,
    HasStreamExternalOutput,
    HasDisable,
    HasSpark,
):
    _action_file_name_conf_key = "spark.stream"
    _run_method_name = "stream"

    def __init__(self) -> None:
        HasSpark.__init__(self)
        HasVars.__init__(self)
        HasStreamExternalInputs.__init__(self)
        HasStreamExternalOutput.__init__(self)
        HasStreamInputs.__init__(self)
        HasStreamOutput.__init__(self)
        HasDisable.__init__(self)
        self.stream = None



from abc import ABC
from argparse import ArgumentParser

from karadoc.common.commands.command import Command


class ReadFromOption(Command, ABC):
    @staticmethod
    def add_arguments(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--read-from",
            dest="remote_env",
            metavar="ENV",
            help="Read data from a remote warehouse in env ENV instead of the local warehouse",
        )



from abc import ABC
from argparse import ArgumentParser

from karadoc.common.commands.command import Command


class DryOption(Command, ABC):
    @staticmethod
    def add_arguments(parser: ArgumentParser) -> None:
        parser.add_argument("--dry", dest="dry", default=False, action="store_true", help="Perform a dry-run")



import importlib
import logging
import sys
from logging import Formatter, StreamHandler
from typing import Optional

from karadoc.common.class_utils import find_class_from_module
from karadoc.common.conf import ConfBox
from karadoc.common.observability import LogEvent
from karadoc.common.observability.console_event import ConsoleEvent
from karadoc.common.observability.logging_utils import add_logging_level

VERBOSE = False
DEFAULT_CONSOLE_LOG_LEVEL = "USER_OUTPUT"
DEFAULT_LOG_LEVEL = "INFO"
DEFAULT_LOG_FORMAT = "%(asctime)s [%(levelname)s] %(name)s - %(message)s"
DEFAULT_VERBOSITY = False


class SimpleConsoleFormatter(Formatter):
    """Formatter used by the stdout_handler to format ConsoleEvents in a more user-friendly way."""

    def __init__(self, fmt=None, datefmt=None, style="%", verbose: bool = DEFAULT_VERBOSITY):
        super().__init__(fmt, datefmt, style)
        self.verbose = verbose

    def format_exc_and_stack(self, record: logging.LogRecord):
        s = ""
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + record.exc_text
        if record.stack_info:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + self.formatStack(record.stack_info)
        return s

    def format(self, record: logging.LogRecord):
        if isinstance(record.msg, ConsoleEvent) and not (self.verbose or VERBOSE):
            console_event = record.msg
            s = console_event.message
            if console_event.extra_console_message:
                s += " " + console_event.extra_console_message
            s += self.format_exc_and_stack(record)
            return s
        elif isinstance(record.msg, LogEvent) and not (self.verbose or VERBOSE):
            return record.msg.message
        else:
            return super(SimpleConsoleFormatter, self).format(record)


def __clear_log_handlers():
    """Remove all log handlers"""
    logging_handlers = list(logging.root.handlers)
    for lh in logging_handlers:
        logging.root.removeHandler(lh)
        lh.close()


def __add_stdout_handler(observability_conf: Optional[ConfBox]):
    """Add handler to write logs on stdout"""
    level = DEFAULT_CONSOLE_LOG_LEVEL
    verbose = DEFAULT_VERBOSITY
    if observability_conf is not None:
        level = observability_conf.get("level", DEFAULT_CONSOLE_LOG_LEVEL).upper()
        verbose = observability_conf.get("verbose", DEFAULT_VERBOSITY)

    stdout_handler = StreamHandler(stream=sys.stdout)

    stdout_handler.setFormatter(SimpleConsoleFormatter(fmt=DEFAULT_LOG_FORMAT, verbose=verbose))

    logging.root.addHandler(stdout_handler)
    stdout_handler.setLevel(DEFAULT_CONSOLE_LOG_LEVEL)
    stdout_handler.setLevel(level)


def _load_observability_handler(handler_conf: ConfBox):
    handler_module = handler_conf.get("type")
    module = importlib.import_module(handler_module)
    handler_class = find_class_from_module(module, logging.Handler)
    azure_handler: logging.Handler = handler_class(**{k: v for k, v in handler_conf.items()})
    level = handler_conf.get("level", DEFAULT_LOG_LEVEL).upper()
    azure_handler.setLevel(level)
    logging.root.addHandler(azure_handler)


def _load_observability_handlers(observability_conf: ConfBox):
    if observability_conf is not None:
        for handler_name, handler_conf in observability_conf.items():
            _load_observability_handler(handler_conf)


def __apply_conf(observability_conf: Optional[ConfBox]) -> None:
    """Apply an observability configuration, if provided"""
    if observability_conf is not None:
        level = observability_conf.get("level", DEFAULT_LOG_LEVEL).upper()
        logging.root.setLevel(level)
        _load_observability_handlers(observability_conf.get("handler"))


def setup_logging(observability_conf: Optional[ConfBox], verbose: Optional[bool]) -> None:
    """Initialize logging with the given configuration.
    We add a new logging level called USER_OUTPUT to avoid flooding the user with INFO-level logs.

    :param observability_conf: dynaconf section related to observability
    :param verbose: Activates verbose mode
    """
    global VERBOSE
    if verbose is not None:
        VERBOSE = verbose
    add_logging_level(level_name="USER_OUTPUT", level_num=logging.INFO + 5)
    __clear_log_handlers()
    __add_stdout_handler(observability_conf)
    logging.root.setLevel(DEFAULT_LOG_LEVEL)
    __apply_conf(observability_conf)



from typing import Dict, Iterable, List, Set

import networkx as nx
from airflow.models import BaseOperator, DagBag

from karadoc.airflow.dag_graph import task_full_name
from karadoc.common.model.table_index import Populate, Table
from karadoc.common.validations import ValidationResult, ValidationSeverity


def is_karadoc_run_task(task: BaseOperator):
    return task.task_type == "KaradocSubmitRunOperator"


def validate_dependencies(
    task: BaseOperator,
    inputs: List[str],
    transitive_closure_graph: nx.DiGraph,
    scheduled_tables: Set[str],
    table_index: Dict[str, Table],
    dag_bag: DagBag,
) -> List[ValidationResult]:
    """Verifies that the dependencies of the Populate are respected by the DAG.
    For that we check that each input is a parent of this table in the transitive closure of the DAG.

    :param task: The airflow task currently being inspected
    :param inputs: List of input table full names declared in the POPULATE.py (job.inputs)
    :param transitive_closure_graph: A dag where the parents of each task are all its ancestors
    :param scheduled_tables: The list of all tables that scheduled in Airflow
    :param table_index: A dict (full_table_name -> table information)
    :param dag_bag: Airflow's DagBag that contains the list of dags.
    :return: a list of ValidationResults
    """

    def task_id_list_to_table_set(nodes):
        def gen():
            for n in nodes:
                dag_id, task_id = n.split(".", maxsplit=1)
                task = dag_bag.dags[dag_id].get_task(task_id)
                if is_karadoc_run_task(task):
                    yield task.table

        return set(gen())

    active_inputs = [input for input in inputs if not table_index[input].populate.disabled]
    ancestor_nodes = list(transitive_closure_graph.predecessors(task_full_name(task)))
    ancestor_tables = task_id_list_to_table_set(ancestor_nodes)
    same_dag_tables = [task.table for task in task.dag.tasks if is_karadoc_run_task(task)]

    res = []
    for input in active_inputs:
        if input not in scheduled_tables:
            msg = "DAG {dag} : Table {table} : the input {input} does not appear in any DAG ".format(
                dag=task.dag_id, table=task.table, input=input
            )
            res.append(ValidationResult("karadoc.airflow.dependency_not_in_any_dag", msg, ValidationSeverity.Error))
        elif input in same_dag_tables and input not in ancestor_tables:
            msg = "DAG {dag} : Table {table} : the input {input} appears in the same DAG but is not upstream ".format(
                dag=task.dag_id, table=task.table, input=input
            )
            res.append(
                ValidationResult(
                    "karadoc.airflow.dependency_in_same_dag_but_not_upstream", msg, ValidationSeverity.Error
                )
            )
        elif input in scheduled_tables and input not in ancestor_tables:
            msg = "DAG {dag} : Table {table} : the input {input} does not appear in any upstream DAG ".format(
                dag=task.dag_id, table=task.table, input=input
            )
            res.append(ValidationResult("karadoc.airflow.dependency_not_upstream", msg, ValidationSeverity.Warning))
    return res


def validate_vars(task, populate: Populate) -> List[ValidationResult]:
    """Check that the there is no discrepancy between the vars passed to the KararadocOperator and the vars
    declared in the populate file.
    """
    res = []
    if task.vars and not populate.vars:
        msg = "DAG {dag} : Table {table} requires no vars but is given {given_vars}".format(
            dag=task.dag_id, table=task.table, given_vars=list(task.vars.keys())
        )
        result = ValidationResult("karadoc.airflow.unused_vars", msg, ValidationSeverity.Warning)
        res.append(result)
    if populate.vars and not task.vars:
        msg = "DAG {dag} : Table {table} requires vars {required_vars} but none is given".format(
            dag=task.dag_id, table=task.table, required_vars=list(populate.vars.keys())
        )
        result = ValidationResult("karadoc.airflow.missing_vars", msg, ValidationSeverity.Error)
        res.append(result)
    if task.vars and populate.vars and task.vars.keys() != populate.vars.keys():
        msg = "DAG {dag} : Table {table} requires vars {required_vars} but is given {given_vars}".format(
            dag=task.dag_id,
            table=task.table,
            required_vars=list(populate.vars.keys()),
            given_vars=list(task.vars.keys()),
        )
        result = ValidationResult("karadoc.airflow.vars_mismatch", msg, ValidationSeverity.Error)
        res.append(result)
    return res


def validate_karadoc_table(task: BaseOperator, table_index, **kwargs) -> List[ValidationResult]:
    """Perform various verifications on Karadoc Tables

    We verify that the Populate file exists, that it is not disabled
    and that the inputs of the table are located upstream in the Airflow DAGs.
    """
    res = []
    if is_karadoc_run_task(task):
        table = table_index.get(task.table)
        if table is None:
            msg = "DAG {dag} : Table {table} does not exists".format(dag=task.dag_id, table=task.table)
            result = ValidationResult("karadoc.table.missing", msg, ValidationSeverity.Error)
            return [result]
        if table.populate is None:
            msg = "DAG {dag} : Table {table} has no Populate".format(dag=task.dag_id, table=task.table)
            result = ValidationResult("karadoc.table.missing_populate", msg, ValidationSeverity.Error)
            return [result]
        if table.populate.disabled:
            msg = "DAG {dag} : Table {table} is disabled".format(dag=task.dag_id, table=task.table)
            result = ValidationResult("karadoc.table.disabled", msg, ValidationSeverity.Error)
            return [result]

        res += validate_vars(task, table.populate)
        res += validate_dependencies(task, table.populate.input_tables, table_index=table_index, **kwargs)

    return res


validations = [
    validate_karadoc_table,
]
"""List of validation methods that will be called for each Airflow Task"""


def validate_task(task: BaseOperator, **kwargs) -> Iterable[ValidationResult]:
    for validation in validations:
        yield from validation(task, **kwargs)


def validate_all_tasks(
    all_tasks: Iterable[BaseOperator],
    dag_bag: DagBag,
    scheduled_tables: Set[str],
    table_index: Dict[str, Table],
    transitive_closure_graph: nx.DiGraph,
) -> Iterable[ValidationResult]:
    for task in all_tasks:
        yield from validate_task(
            task,
            dag_bag=dag_bag,
            transitive_closure_graph=transitive_closure_graph,
            scheduled_tables=scheduled_tables,
            table_index=table_index,
        )



import threading
from typing import Optional

from pyspark.sql import SparkSession

from karadoc.common.exceptions import IllegalJobInitError
from karadoc.spark.utils import get_spark_session


def _partition_to_path(partition) -> str:
    static_partitions = []
    if partition:
        for element in partition:
            if element and type(element) == tuple:
                static_partitions.append(element)

    if len(static_partitions) > 0:
        return "/" + "/".join(["%s=%s" % (p_key, p_value) for (p_key, p_value) in static_partitions])
    else:
        return ""


def _table_name_to_hdfs_path(base_path, schema_name: str, table_name: str, partition) -> str:
    partition_path = _partition_to_path(partition)
    return "%s/%s.db/%s%s" % (base_path, schema_name, table_name, partition_path)


class HasSpark:
    """Before loading a POPULATE, this global variable must be set to True to prevent users
    from accidentally using the SparkContext outside of the `run` method.
    """

    _global_init_lock = False

    def __init__(self) -> None:
        # Attributes that the user is not supposed to change
        self.__spark: Optional[SparkSession] = None
        self.__threadlocal_spark = threading.local()

    def init(self, app_name: str = None, spark_conf: dict = None):
        if HasSpark._global_init_lock:
            raise IllegalJobInitError("Error: the spark context should not be initialized in an action file.")
        self.__spark = get_spark_session(app_name, spark_conf)

    def detach_spark_session(self) -> None:
        """Replace the SparkSession of this job with a new separate SparkSession that has separate SQLConf,
        registered temporary views and UDFs, but shared SparkContext and table cache.
        :return:
        """
        self.__threadlocal_spark.val = self.__spark.newSession()

    def reset_spark_session(self) -> None:
        """"""
        self.spark.sql("RESET")

    def reset(self) -> None:
        self.__init__()

    @property
    def spark(self) -> Optional[SparkSession]:
        if hasattr(self.__threadlocal_spark, "val"):
            return self.__threadlocal_spark.val
        else:
            return self.__spark



import os
from math import ceil
from typing import List

from pyspark.sql import DataFrame
from pyspark.sql import functions as f
from pyspark.sql.types import StringType

FEATURES = ["val_count", "val_frequency"]
COLOR_PAL = [
    {"sunset_orange": "#ff5050"},
    {"gray": "#828282"},
    {"aqua_forest": "#5FA364"},
    {"blue_bayoux": "#4C5C6F"},
    {"pacific_blue": "#00A6BC"},
    {"flamingo": "#F26533"},
    {"mondo": "#504136"},
    {"malta": "#BCAF9C"},
    {"double_colonial_white": "#EEE8AA"},
    {"orange_peel": "#E69F00"},
    {"cornflower_blue": "#6495ED"},
    {"olive": "#808000"},
]


def __null_fields_to_missing_or_other(df: DataFrame) -> DataFrame:
    """Replace all null values in columns with the string 'MISSING' or 'OTHER'
    depending on the value of 'is_other' boolean.
    :param df: a DataFrame
    :return:
    """
    df_res = df
    data_types = ["string"]
    data_cols = [col_name for col_name, type in df.dtypes if type in data_types]
    for column in data_cols:
        df_res = df_res.withColumn(
            column,
            f.when(((f.col(column).isNull()) & (f.col("is_other") == "false")), f.lit("MISSING"))
            .when(((f.col(column).isNull()) & (f.col("is_other") == "true")), f.lit("OTHER_BUCKET"))
            .otherwise(f.col(column)),
        )
    return df_res


def __configure_draw(df: DataFrame, cols: List[str]):
    """
    Create a dictionary with the constant elements (figure, dataset, status list...) used to draw the timeline
    :param df: A Dataframe
    :param cols: The name of the columns to use as groups for the analysis
    :return: A dictionary with a Pandas dataframe, a figure (Figure object) and its axes (Axes object), the number
        of distinct keys used, a list of distinct status if it exists: [dataset, fig, axs, nb_keys, status_list]
    """
    if cols[1] == "":
        has_status_cohort = False
        dataset = (
            df.withColumnRenamed(cols[0], "cohort")
            .withColumn("status", f.lit(None).cast(StringType()))
            .withColumn("val_frequency", f.round(f.col("val_frequency"), 2))
            .orderBy("cohort", "val_count")
            .toPandas()
        )
        status_list = [""]
        keys = dataset["key"].unique()
        figsize = (3 * len(keys) + 11, 3 * len(keys) + 2)
        return dataset, figsize, keys, status_list, has_status_cohort
    else:
        has_status_cohort = True
        dataset = (
            df.withColumnRenamed(cols[0], "cohort")
            .withColumnRenamed(cols[1], "status")
            .withColumn("val_frequency", f.round(f.col("val_frequency"), 2))
            .orderBy("cohort", "val_count")
            .toPandas()
        )
        status_list = dataset["status"].unique()
        keys = dataset["key"].unique()
        figsize = (80, 3 * len(keys) + 2)
        return dataset, figsize, keys, status_list, has_status_cohort


def __get_sorted_vals(dataset, key):
    vals = dataset[dataset.key == key]["val"].unique()
    sorted_vals = []
    if "MISSING" in vals:
        sorted_vals += ["MISSING"]
    if "OTHER_BUCKET" in vals:
        sorted_vals += ["OTHER_BUCKET"]
    sorted_vals += [i for i in vals if i not in {"MISSING", "OTHER_BUCKET"}]
    return sorted_vals


def draw(df: DataFrame, cohort_cols: List[str], filepath: str, filename: str):  # NOSONAR
    """
    Draw a stackplot with the given dataframe into a .png file with the val_count draw (left) and the frequency draw
    (right) for the given cohorts
    :param df: A Dataframe
    :param cohort_cols: The name of the columns to use as groups for the analysis
    :param filepath: The path of the .png file created
    :param filename: The name of the .png file created
    :return:
    """
    import matplotlib.pyplot as plt
    import numpy as np
    from matplotlib import use
    from pandas.plotting import register_matplotlib_converters

    register_matplotlib_converters()
    use("Agg")

    df_analysed = __null_fields_to_missing_or_other(df)
    dataset, figsize, keys, status_list, has_status_cohort = __configure_draw(df_analysed, cohort_cols)
    nb_keys = len(keys)
    fig, axs = plt.subplots(nb_keys, 2 * len(status_list), figsize=figsize)
    x_axis, legend = None, None
    for pos, key in enumerate(keys):
        nb = 0
        for nb_status, status in enumerate(status_list):
            for nb_tables, feature in enumerate(FEATURES):
                p = []
                labels = []
                nb_tables = nb_tables + nb_status + nb
                sorted_vals = __get_sorted_vals(dataset, key)
                for i, val_local in enumerate(sorted_vals):
                    row_filter = (dataset.val == val_local) & (dataset.key == key)
                    if has_status_cohort:
                        row_filter &= dataset.status == status
                    if i == 0:
                        x_axis = np.array(dataset[row_filter]["cohort"])
                        p.append(x_axis)
                    p.append(np.array(dataset[row_filter][feature]))
                    labels.append(val_local)
                custom_axs = axs[nb_tables] if nb_keys <= 1 else axs[pos][nb_tables]
                custom_axs.stackplot(*p, colors=[list(a.values())[0] for a in COLOR_PAL], alpha=0.6)
                custom_axs.set_xticks(x_axis[:: int(ceil(0.086 * len(x_axis) + 0.091))])
                custom_axs.margins(x=0, y=0)
                plt.setp(custom_axs.xaxis.get_majorticklabels(), rotation=45, fontsize="x-small")
                custom_axs.set_title(feature + " " + key) if cohort_cols[1] == "" else custom_axs.set_title(
                    feature + " " + key + " for " + status
                )
                custom_axs.set_ylabel(feature)
                legend = custom_axs.legend(labels=labels, ncol=1, loc=2, bbox_to_anchor=(1, 1.01), prop={"size": 6.5})
            nb += 1
    plt.tight_layout()
    if not os.path.isdir(filepath):
        os.makedirs(filepath)
    plt.savefig(filepath + "/" + filename + ".png", bbox_extra_artists=(legend,), bbox_inches="tight", dpi=180)
    plt.close()



from typing import Any, Dict, Iterable, Optional, Tuple

from dynaconf.utils.boxing import DynaBox

from karadoc.common.conf import keyvault
from karadoc.common.utils.assert_utils import assert_true


class ConfBox:
    """Wrapper Class of Dynabox that allows to fetch values from secure vaults like Azure Keyvault"""

    def __init__(self, box: DynaBox, env: Optional[str] = None):
        """

        :param box: The DynaBox to wrap
        :param env: The env this ConfBox was built in
        """
        assert_true(isinstance(box, DynaBox))
        self.__box = box
        self.__env = env

    def __fetch_if_secret(self, value: Any) -> Any:
        if isinstance(value, dict) and "secret" in value:
            secret_env, secret_path = list(value.get("secret").items())[0]
            return keyvault.get_secret(conn_name=secret_env, secret_id=secret_path, env=self.__env)
        elif isinstance(value, DynaBox):
            return ConfBox(value, self.__env)
        else:
            return value

    def is_secret(self, item):
        value = self.__box.get(item)
        return isinstance(value, dict) and "secret" in value

    def __getattr__(self, item, *args, **kwargs):
        return self.__fetch_if_secret(self.__box.__getattr__(item, *args, **kwargs))

    def __getitem__(self, item, *args, **kwargs):
        return self.__fetch_if_secret(self.__box.__getitem__(item, *args, **kwargs))

    def __contains__(self, item):
        return self.__box.__contains__(item)

    def __getstate__(self):
        """Function used by pickle to serialize the class"""
        return self.__box

    def __setstate__(self, state):
        """Function used by pickle to deserialize the class"""
        self.__init__(state)

    def keys(self):
        return self.__box.keys()

    def items(self):
        for k, v in self.__box.items():
            yield k, self.__fetch_if_secret(v)

    def values(self):
        for k, v in self.__box.values():
            yield self.__fetch_if_secret(v)

    def get(self, item, default=None, *args, **kwargs):
        return self.__fetch_if_secret(self.__box.get(item, default, *args, **kwargs))

    def __str__(self) -> str:
        return str(self.__box.to_dict())

    def __repr__(self) -> str:
        return f"<ConfBox({repr(self.__box.to_dict())})>"

    def __eq__(self, other) -> bool:
        if isinstance(other, ConfBox):
            return self.__box == other.__box
        else:
            return NotImplemented

    def __to_flat_dict(self, prefix: str = "") -> Iterable[Tuple[str, object]]:
        for k, v in self.items():
            new_prefix = prefix
            if new_prefix != "":
                new_prefix = new_prefix + "."
            new_prefix += str(k)
            if isinstance(v, ConfBox):
                yield from v.__to_flat_dict(new_prefix)
            else:
                yield new_prefix, v

    def to_flat_dict(self) -> Dict[str, object]:
        """Returns a flat dictionary with all the {key: value} settings contained into this ConfBox

        Example:

        >>> conf = dict_to_dynabox({
        ...    "spark.driver.cores": 2,
        ...    "spark.driver.memory": "4g"
        ... })
        >>> conf_box = ConfBox(conf)
        >>> conf_box
        <ConfBox({'spark': {'driver': {'cores': 2, 'memory': '4g'}}})>
        >>> conf_box.to_flat_dict()
        {'spark.driver.cores': 2, 'spark.driver.memory': '4g'}

        """
        return {k: v for (k, v) in self.__to_flat_dict()}


def __unflatten_dict(d: Dict[str, Any]) -> Dict[str, Any]:
    """Transform a flat dictionary with keys containing dots (".") into a nested dictionary with keys split by "."

    >>> __unflatten_dict({'a': 'b'})
    {'a': 'b'}
    >>> __unflatten_dict({'a.b': 'c'})
    {'a': {'b': 'c'}}
    >>> __unflatten_dict({'a.b.c': 'd'})
    {'a': {'b': {'c': 'd'}}}
    >>> __unflatten_dict({'a': {'b.c': 'd'}})
    {'a': {'b': {'c': 'd'}}}

    :param d: a flat dict
    :return: a nested dict
    """
    res = {}
    for key, value in d.items():
        if isinstance(value, dict):
            value = __unflatten_dict(value)
        sub_keys = key.split(".")
        current = res
        for sub_key in sub_keys[:-1]:
            if sub_key not in current:
                current[sub_key] = {}
            current = current[sub_key]
        current[sub_keys[-1]] = value
    return res


def dict_to_dynabox(d):
    """Transforms a dict into a DynaBox. Mostly used for testing.

    >>> dict_to_dynabox({"a": {"b" : "c"}})
    <Box: {'a': {'b': 'c'}}>
    >>> dict_to_dynabox({"a": {"b" : "c"}}).get("a")
    <Box: {'b': 'c'}>
    >>> dict_to_dynabox({"a.b" : "c"})
    <Box: {'a': {'b': 'c'}}>
    >>> dict_to_dynabox({"a.b" : "c"}).get("a")
    <Box: {'b': 'c'}>

    :param d:
    :return:
    """
    if isinstance(d, dict):
        return DynaBox({k: dict_to_dynabox(v) for k, v in __unflatten_dict(d).items()})
    else:
        return d



from karadoc.common.job_core.has_batch_inputs import HasBatchInputs
from karadoc.common.job_core.has_batch_output import HasBatchOutput
from karadoc.common.job_core.has_disable import HasDisable
from karadoc.common.job_core.has_external_inputs import HasExternalInputs
from karadoc.common.job_core.has_external_outputs import HasExternalOutputs
from karadoc.common.job_core.has_keys import HasKeys
from karadoc.common.job_core.has_spark import HasSpark
from karadoc.common.job_core.has_vars import HasVars


class SparkBatchJob(
    HasBatchInputs,
    HasBatchOutput,
    HasVars,
    HasExternalInputs,
    HasExternalOutputs,
    HasKeys,
    HasDisable,
    HasSpark,
):
    _action_file_name_conf_key = "spark.batch"
    _run_method_name = "run"

    def __init__(self) -> None:
        HasSpark.__init__(self)
        HasBatchInputs.__init__(self)
        HasBatchOutput.__init__(self)
        HasVars.__init__(self)
        HasExternalInputs.__init__(self)
        HasExternalOutputs.__init__(self)
        HasKeys.__init__(self)
        HasDisable.__init__(self)

        self.run = None



from enum import IntEnum
from typing import Callable


def _bin_op(op_name: str) -> Callable[["ReturnCode", "ReturnCode"], "ReturnCode"]:
    """Create a method for given binary operator"""

    def func(self: "ReturnCode", other: "ReturnCode") -> "ReturnCode":
        if not isinstance(other, ReturnCode):
            return NotImplemented
        op = getattr(self.value, op_name)
        return ReturnCode(op(other.value))

    return func


class ReturnCode(IntEnum):
    """Enum return code for a karadoc command.

    It's either Success (0) or Error (1)
    It supports & and | operators

    Examples:

    >>> ReturnCode.Success != ReturnCode.Error
    True
    >>> ReturnCode.Success == ReturnCode(0)
    True
    >>> ReturnCode.Error == ReturnCode(1)
    True
    >>> ReturnCode.Success & ReturnCode.Error
    <ReturnCode.Success: 0>
    >>> ReturnCode.Success | ReturnCode.Error
    <ReturnCode.Error: 1>

    """

    Success = 0
    Error = 1

    __or__ = _bin_op("__or__")
    __and__ = _bin_op("__and__")



from typing import List

from karadoc.common.job_core.has_batch_inputs import HasBatchInputs
from karadoc.common.job_core.has_before_after import HasBeforeAfter
from karadoc.common.job_core.has_disable import HasDisable
from karadoc.common.job_core.has_external_inputs import HasExternalInputs
from karadoc.common.job_core.has_spark import HasSpark
from karadoc.common.job_core.has_vars import HasVars
from karadoc.common.quality.checks import Alert, Metric
from karadoc.common.utils.assert_utils import assert_true


class QualityCheckJob(HasBatchInputs, HasVars, HasBeforeAfter, HasExternalInputs, HasDisable, HasSpark):
    _action_file_name_conf_key = "spark.quality_check"

    def __init__(self) -> None:
        HasSpark.__init__(self)
        HasBatchInputs.__init__(self)
        HasVars.__init__(self)
        HasBeforeAfter.__init__(self)
        HasExternalInputs.__init__(self)
        HasDisable.__init__(self)

        self.__alerts: List[Alert] = []
        self.__metrics: List[Metric] = []

    @property
    def alerts(self) -> List[Alert]:
        return sorted(list(self.__alerts), key=lambda check: check._creation_rank)

    def add_alert(self, alert: Alert) -> None:
        assert_true(isinstance(alert, Alert))
        self.__alerts.append(alert)

    @property
    def metrics(self) -> List[Metric]:
        return sorted(list(self.__metrics), key=lambda check: check._creation_rank)

    def add_metric(self, metric: Metric) -> None:
        assert_true(isinstance(metric, Metric))
        self.__metrics.append(metric)



import os
import shutil
from argparse import ArgumentParser, Namespace
from pathlib import Path

from karadoc.common import conf
from karadoc.common.commands.command import Command
from karadoc.common.utils.assert_utils import assert_true


def __pull_from_azure_blob(args, remote_env, schema_name, table_name):
    from azure.storage.blob import BlockBlobService
    from azure.storage.blob.models import Blob

    # tqdm is a nice library that wraps a list and automagically displays a progress bar
    from tqdm import tqdm

    storage_account = remote_env["storage_account"]
    storage_key = remote_env["storage_key"]
    container_name = remote_env["container_name"]
    warehouse_dir = remote_env["warehouse_dir"]

    blob_service = BlockBlobService(account_name=storage_account, account_key=storage_key)

    source_prefix = warehouse_dir
    table_relative_path = "%s.db/%s" % (schema_name, table_name)
    source_full_path = "%s/%s" % (source_prefix, table_relative_path)
    local_warehouse = conf.get_warehouse_folder_location()
    output_folder = "%s/%s" % (local_warehouse, table_relative_path)

    print(
        "input: %s" % ("wasbs://%s@%s.blob.core.windows.net/%s" % (container_name, storage_account, source_full_path))
    )
    print("output: %s" % output_folder)

    all_blobs = blob_service.list_blobs(container_name=container_name, prefix=source_full_path)
    file_blobs = [f for f in all_blobs if f.properties.content_length > 0]
    nb_files = len(file_blobs)

    if args.dry:
        print("dry run: would copy %s files" % nb_files)
    else:
        # we delete the destination folder first
        if os.path.exists(output_folder) and os.path.isdir(output_folder):
            shutil.rmtree(output_folder)

        f: Blob
        for f in tqdm(file_blobs):
            relative_path = f.name[len(source_prefix):]  # fmt: skip
            dest = "%s/%s" % (local_warehouse, relative_path)

            # create the destination directory first if it does not exist
            if not os.path.exists(str(Path(dest).parent)):
                os.makedirs(str(Path(dest).parent))

            # download the file
            blob_service.get_blob_to_path(container_name=container_name, blob_name=f.name, file_path=dest)

        print("copied %s files" % len(file_blobs))


def pull_table(args, table):
    print("pulling table %s from %s" % (table, args.from_env))
    a = table.split(".")
    schema_name = a[0]
    table_name = a[1] if len(a) > 1 else ""
    remote_env = conf.get_connection_conf(args.from_env)
    assert_true(
        (remote_env["type"] == "wasbs"),
        f"Remote environment {args.from_env} is of type {remote_env['type']}, which is not supported yet",
    )
    __pull_from_azure_blob(args, remote_env, schema_name, table_name)


class PullCommand(Command):
    description = "pull data from a remote environment to the local environment"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        parser.add_argument("--dry", dest="dry", default=False, action="store_true", help="perform a dry-run")
        parser.add_argument("--tables", metavar="table", type=str, nargs="+", help="Names of the tables to pull")
        parser.add_argument(
            "--from",
            metavar="env",
            dest="from_env",
            required=True,
            type=str,
            help="Name of the remote environment to pull from",
        )

    @staticmethod
    def do_command(args: Namespace):
        for table in args.tables:
            pull_table(args, table)



from abc import ABC
from argparse import ArgumentParser

from karadoc.common.commands.command import Command


class TablesOption(Command, ABC):
    @staticmethod
    def add_arguments(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--tables", dest="tables", metavar="table", type=str, nargs="+", required=True, help="Names of the tables"
        )



from typing import Generator

import networkx as nx

from karadoc.common.validations import (
    ValidationResult,
    ValidationResultTemplate,
    ValidationSeverity,
)

ValidationResult_GraphContainsCycle = ValidationResultTemplate(
    check_type="karadoc.graph.contains_cycle",
    message_template='Dependency cycle found : {cycle}"',
    default_severity=ValidationSeverity.Critical,
)


def validate_graph(graph: nx.DiGraph) -> Generator[ValidationResult, None, None]:
    if not nx.is_directed_acyclic_graph(graph):
        yield ValidationResult_GraphContainsCycle(cycle=nx.find_cycle(graph))



import logging
from unittest import TestCase

from karadoc.spark.utils import get_spark_session


class PySparkTest(TestCase):
    @classmethod
    def suppress_py4j_logging(cls):
        logger = logging.getLogger("py4j")
        logger.setLevel(logging.WARN)

    @classmethod
    def __create_testing_pyspark_session(cls):
        return get_spark_session()

    @classmethod
    def setUpClass(cls):
        cls.suppress_py4j_logging()
        cls.spark = cls.__create_testing_pyspark_session()

    @classmethod
    def tearDownClass(cls):
        cls.spark.stop()



import sys
from argparse import ArgumentParser, Namespace

from pyspark.sql import DataFrame

from karadoc.common import conf, stream
from karadoc.common.commands.command import Command
from karadoc.common.commands.options.dry_option import DryOption
from karadoc.common.commands.options.spark_conf_option import SparkConfOption
from karadoc.common.commands.options.tables_option import TablesOption
from karadoc.common.commands.options.vars_option import VarsOption
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.commands.utils import run_job_with_logging
from karadoc.common.exceptions import JobDisabledException
from karadoc.common.model import variables
from karadoc.common.stream.spark_stream_job import SparkStreamJob
from karadoc.common.validations import fail_if_results
from karadoc.common.validations.connection_validations import validate_connections
from karadoc.common.validations.job_validations import (
    validate_inputs,
    validate_output_partition,
)


def inspect_df(df: DataFrame):
    """Function used for unit tests. Mock this function to be able to inspect the produced DataFrame
    during the execution of the command."""
    pass


def inspect_job(job: SparkStreamJob):
    """Function used for unit tests. Mock this function to be able to inspect the SparkStreamJob object
    during the execution of the command."""
    pass


def _run_job(args: Namespace, job: SparkStreamJob, **kwargs):
    stream.exec.print_job_properties(job)
    fail_if_results(validate_connections(job))
    fail_if_results(validate_inputs(job))
    fail_if_results(validate_output_partition(job))
    if job.disable:
        if not args.dry:
            raise JobDisabledException()
        else:
            print("WARN: The job has been disabled, it should not be launched")
    if args.batch_interval is not None and args.streaming_mode != "microbatch":
        print("ERROR: batch_interval is only specified when streaming mode is microbatch")
        sys.exit(1)
    if len(args.tables) > 1:
        print("ERROR: multiple tables are not supported for the moment for --tables argument")
        sys.exit(1)
    if not args.dry:
        job.init(app_name=args.raw_args, spark_conf=args.spark_conf)
        df = job.stream()
        inspect_df(df)
        stream.exec.write_stream_output(job, df, args.no_export, args.streaming_mode, args.batch_interval)
    elif not conf.is_dev_env():
        job.init()
    inspect_job(job)


class StreamCommand(Command):
    description = "run STREAM.py scripts for the specified tables"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        TablesOption.add_arguments(parser)
        DryOption.add_arguments(parser)
        VarsOption.add_arguments(parser)
        SparkConfOption.add_arguments(parser)

        parser.add_argument(
            "--no-export", dest="no_export", default=False, action="store_true", help="Disable exporting data"
        )
        parser.add_argument(
            "--streaming-mode",
            dest="streaming_mode",
            default="microbatch",
            type=str,
            choices=["once", "microbatch"],
            help="(Default: microbatch) Define the timing of streaming data processing, whether the query is going "
            "to be executed as micro-batch query with a fixed batch interval or as a one batch query.",
        )
        parser.add_argument(
            "--batch-interval",
            dest="batch_interval",
            type=str,
            help="The query will be executed with micro-batches mode, "
            "where micro-batches will be kicked off at the user-specified intervals"
            "batch-interval time interval is a string, e.g. '5 seconds', '1 minute'.",
        )

    @staticmethod
    def do_command(args: Namespace):
        return_code = ReturnCode.Success
        vars_list = variables.expand_vars(args.vars)
        jobs = [
            (table, job_vars, stream.exec.load_runnable_stream_file(table, job_vars))
            for table in args.tables
            for job_vars in vars_list
        ]
        for model_id, job_vars, job in jobs:
            return_code &= run_job_with_logging(
                func=_run_job, job_type=SparkStreamJob, model_id=model_id, job_vars=job_vars, args=args, job=job
            )
        return return_code



from pyspark.sql import DataFrame
from pyspark.sql.streaming import DataStreamWriter

from karadoc.common.job_core.has_output import HasOutput
from karadoc.spark.conf import (
    get_stream_default_output_format,
    get_write_options_for_format,
)


class HasStreamOutput(HasOutput):
    def __init__(self) -> None:
        super().__init__()
        self.output_format = get_stream_default_output_format()

    def write_table(self, df: DataFrame):
        return self._write_stream_table(
            stream_df=df,
            path=self.hdfs_output(),
            output_format=self.output_format,
            output_options=self.output_options,
            partitions=self.dynamic_partitions(),
        )

    @staticmethod
    def _write_stream_table(
        stream_df: DataFrame, path: str, output_format: str, output_options=None, partitions=None
    ) -> DataStreamWriter:
        if output_options is None:
            output_options = {}
        if partitions is None:
            partitions = []

        write_options = get_write_options_for_format(output_format)
        options = {**write_options, **output_options}

        return (
            stream_df.writeStream.partitionBy(partitions).option("path", path).options(**options).format(output_format)
        )



from typing import Dict, List, Set

from airflow.models import BaseOperator, DagBag

from karadoc.airflow.dag_validations import validate_all_dags
from karadoc.airflow.task_validations import is_karadoc_run_task, validate_all_tasks
from karadoc.common.model.table_index import Table
from karadoc.common.validations import ValidationResult, ValidationSeverity


def list_tasks(dag_bag: DagBag) -> List[str]:
    return [task for dag in dag_bag.dags.values() for task in dag.tasks]


def list_scheduled_tables(tasks: List[BaseOperator]) -> Set[str]:
    """Return the list of all tasks that correspond to a KaradocSubmitRunOperator"""
    return {task.table for task in tasks if is_karadoc_run_task(task)}


def validate_dag_bag(dag_bag: DagBag, table_index: Dict[str, Table]) -> List[ValidationResult]:
    from karadoc.airflow.dag_graph import build_graph, transitive_closure

    dag_graph = build_graph(dag_bag)
    transitive_closure_graph = transitive_closure(dag_graph)
    all_tasks = list_tasks(dag_bag)
    scheduled_tables = list_scheduled_tables(all_tasks)
    error_messages = [
        ValidationResult(
            "airflow.import.error",
            "could not load file {file} : {message}".format(file=file, message=message),
            ValidationSeverity.Error,
        )
        for file, message in dag_bag.import_errors.items()
    ]
    error_messages += list(
        validate_all_tasks(all_tasks, dag_bag, scheduled_tables, table_index, transitive_closure_graph)
    )
    error_messages += list(validate_all_dags(dag_bag))
    return sorted(error_messages, key=(lambda e: e.severity.value))



from argparse import ArgumentParser, Namespace
from typing import List

import networkx as nx
from pyspark import Row
from pyspark.sql import DataFrame

from karadoc.common.commands.command import Command
from karadoc.common.graph import table_graph
from karadoc.common.graph.table_graph import find_tables_to_disable
from karadoc.common.model import table_index
from karadoc.common.output.local_export import local_export_dataframe
from karadoc.common.validations import fail_if_results
from karadoc.common.validations.graph_validations import validate_graph


def _str_list_to_df(data: List[str]) -> DataFrame:
    """Builds a Spark DataFrame out of a List[str]."""
    from karadoc.common import Job

    job = Job()
    job.init()
    df = job.spark.createDataFrame([Row(t) for t in data], schema="full_table_name STRING")
    df = df.orderBy("full_table_name")
    return df


class FindTablesToDisableCommand(Command):
    description = (
        "(experimental) find tables that can be disabled. A table can be disabled if it "
        "has no external output and if all of its direct successors are already disabled or can be disabled."
    )

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        parser.add_argument(
            "--format",
            dest="format",
            default="txt",
            choices=["txt", "markdown", "csv", "xlsx"],
            help="(Default: txt) Output format for the command results.",
        )
        parser.add_argument(
            "--output",
            dest="output",
            default=None,
            metavar="PATH",
            help="Path of the file where the results of the command will be written. "
            "By default, the results are written on stdout, except for binary format (e.g. xlsx) "
            "for which the default is connections.FORMAT",
        )

    @staticmethod
    def do_command(args: Namespace):
        binary_formats = ["xlsx"]
        if args.output is None and args.format in binary_formats:
            args.output = f"connections.{args.format}"

        index = table_index.build_table_index()
        graph = table_graph.build_graph(index)
        validation_results = validate_graph(graph)

        graph: nx.DiGraph = nx.transitive_reduction(graph)

        tables_to_disable = find_tables_to_disable(index, graph)
        df = _str_list_to_df(tables_to_disable)
        local_export_dataframe(df=df, output=args.output, format=args.format)

        fail_if_results(validation_results)



import time


class Stopwatch:
    """Provides a timer that can be used with a context manager.

    Example:
    > stopwatch = Stopwatch()
    > time.sleep(0.1)
    > print("%0f seconds" % stopwatch.duration_in_seconds)
    0.1 seconds

    """

    def __init__(self) -> None:
        self.start_time = time.time()

    @property
    def duration_in_seconds(self) -> float:
        return round(time.time() - self.start_time, 3)

    @property
    def duration_str(self) -> str:
        return "%0.3f s" % self.duration_in_seconds



import os
import sys
from pathlib import Path
from typing import Optional

from pyspark.sql import SparkSession

from karadoc.common.conf import get_libs_folder_location


def get_spark_session(app_name: Optional[str] = None, extra_spark_conf: Optional[dict] = None) -> SparkSession:
    """Gets an existing :class:`SparkSession` or, if there is no existing one, creates a new one.

    Spark default settings can be set via the `{env}.spark.conf` in `settings.toml`, and can also be overridden
    using this method's optional argument `extra_spark_conf`.

    Example:

    In your settings.toml:

    .. code-block:: python

        # In settings.toml, this setting configures the default driver memory on all environments
        [default.spark.conf]
          spark.driver.memory="4g"

        # While this setting overrides the default driver memory in the "prod" environment
        [prod.spark.conf]
          spark.driver.memory="6g"

    # It can also be overridden via the command line for a specific job like this :
    karadoc run --spark-conf spark.driver.memory=8g ...

    :param app_name:
    :param extra_spark_conf: Spark configuration parameters to over
    :return:
    """
    # TODO: this late import is a temporary workaround for cyclic dependencies
    from karadoc.common import conf
    from karadoc.spark.conf import get_spark_conf

    if extra_spark_conf is None:
        extra_spark_conf = {}
    spark_conf = {**get_spark_conf(), **extra_spark_conf}
    builder: SparkSession.Builder = SparkSession.builder
    if conf.is_dev_env():
        builder = builder.master("local[4]")
    for k, v in spark_conf.items():
        builder = builder.config(k, v)
    if app_name is None:
        app_name = conf.APPLICATION_NAME.lower()
    spark: SparkSession = builder.appName(app_name).getOrCreate()
    _add_libs_folder_to_spark_python_path(spark)

    # As a workaround to this spark issue https://issues.apache.org/jira/browse/SPARK-38870,
    # we reset the builder options
    spark.Builder._options = {}
    return spark


def _add_libs_folder_to_spark_python_path(spark: SparkSession) -> None:
    # This simulates the first time the karadoc module is imported
    libs_folder = str(Path(get_libs_folder_location()).absolute())
    spark_env = spark.sparkContext.environment
    sys.path.extend([libs_folder])
    if "PYTHONPATH" not in spark_env:
        spark_env["PYTHONPATH"] = libs_folder
    else:
        spark_env["PYTHONPATH"] += os.pathsep + libs_folder



from typing import Any


class AnythingClass:
    """Class that will always return true when compared to something.
    Useful for asserting functions calls with the `unittest.mock` library

    WARNING: that 'trick' may not work when compared to any object `obj` such that obj.__eq__(Anything) returns False.
    Objects with a well-implemented equality should return NotImplemented when compared to something
    they don't know, not False.

    - Anything == Other will always return True
    - Other == Anything will return Other.__eq__(Anything) if is implemented, True otherwise

    Example:

    >>> AnythingClass() == ""
    True
    >>> "" == AnythingClass()
    True
    >>> "".__eq__(AnythingClass())
    NotImplemented

    >>> class NothingClass:
    ...    def __eq__(self, other):
    ...        return False
    >>> AnythingClass() == NothingClass()
    True
    >>> NothingClass() == AnythingClass()
    False
    >>> NothingClass().__eq__(AnythingClass())
    False
    """

    def __eq__(self, other: Any) -> bool:
        return True

    def __hash__(self) -> int:
        return 0


Anything = AnythingClass()
"""Static object that will always return true when compared to something.
Useful for asserting functions calls with the `unittest.mock` library.

WARNING: that 'trick' may not work when compared to any object `obj` such that obj.__eq__(Anything) returns False.
Objects with a well-implemented equality should return NotImplemented when compared to something
they don't know, not False.
See the examples in `AnythingClass` for more details.
"""



from typing import Any, Dict, Generator, Iterable

from airflow.models import DagBag

from karadoc.airflow.utils import is_trigger
from karadoc.common.validations import (
    ValidationResult,
    ValidationResultTemplate,
    ValidationSeverity,
)

ValidationResult_DagTriggeredMoreThanOnce = ValidationResultTemplate(
    check_type="karadoc.airflow.dag_triggered_more_than_once",
    message_template="The dag {dag_id} is triggered {count} times",
    default_severity=ValidationSeverity.Warning,
)

ValidationResult_DagNotScheduledNorTriggered = ValidationResultTemplate(
    check_type="karadoc.airflow.dag_not_scheduled_nor_triggered",
    message_template="The dag {dag_id} is never scheduled nor triggered",
    default_severity=ValidationSeverity.Warning,
)

ValidationResult_DagScheduledAndTriggered = ValidationResultTemplate(
    check_type="karadoc.airflow.dag_scheduled_and_triggered",
    message_template="The dag {dag_id} is scheduled and triggered",
    default_severity=ValidationSeverity.Warning,
)


def _element_count(iterable: Iterable[Any]) -> Dict[Any, int]:
    """Count the number of occurrences of each distinct value in a list and returns a Dict[elem, int]

    >>> _element_count([1, 2, 2])
    {1: 1, 2: 2}
    >>> _element_count("abracadabra")
    {'a': 5, 'b': 2, 'r': 2, 'c': 1, 'd': 1}
    """
    res: Dict[Any, int] = dict()
    for item in iterable:
        if item in res:
            res[item] += 1
        else:
            res[item] = 1
    return res


def _get_triggered_dags(dag_bag: DagBag) -> Dict[str, int]:
    """List all dags that are triggered along with the number of times they are triggered

    :return: a Dict(dag_id, number_of_times_triggered)"""

    def list_all() -> Generator[str, None, None]:
        for dag in dag_bag.dags.values():
            for task in dag.tasks:
                if is_trigger(task):
                    yield task.trigger_dag_id

    return _element_count(list_all())


def _get_scheduled_dags(dag_bag: DagBag) -> Generator[str, None, None]:
    """List all dags that are scheduled"""
    for dag in dag_bag.dags.values():
        if dag.schedule_interval is not None:
            yield dag.dag_id


def validate_all_dags(dag_bag: DagBag) -> Generator[ValidationResult, None, None]:
    triggered_dags_with_count = _get_triggered_dags(dag_bag)
    triggered_dags = set(triggered_dags_with_count.keys())
    all_dags = set(dag_bag.dags.keys())
    scheduled_dags = set(_get_scheduled_dags(dag_bag))

    for dag_id, count in triggered_dags_with_count.items():
        if count > 1:
            yield ValidationResult_DagTriggeredMoreThanOnce(dag_id=dag_id, count=count)

    for dag_id in all_dags.difference(triggered_dags).difference(scheduled_dags):
        yield ValidationResult_DagNotScheduledNorTriggered(dag_id=dag_id)

    for dag_id in scheduled_dags.intersection(triggered_dags):
        yield ValidationResult_DagScheduledAndTriggered(dag_id=dag_id)



import re


def strip_margin(text: str) -> str:
    """For every line in this string:
    Strip a leading prefix consisting of blanks or control characters followed by | from the line.

    This function is similar to Scala's stripMargin function.

    >>> print(strip_margin('''
    ...    |a
    ...    |b
    ...    |'''))
    a
    b
    <BLANKLINE>

    """
    s = re.sub(r"\n[ \t]*\|", "\n", text)
    if s.startswith("\n"):
        return s[1:]
    else:
        return s



from typing import Optional

from karadoc.common.utils.assert_utils import assert_true


class JobBase:
    _action_file_name_conf_key: str
    _run_method_name: Optional[str] = None
    output: str

    @classmethod
    def get_action_file_name(cls) -> str:
        from karadoc.common import conf

        assert_true(
            (cls._action_file_name_conf_key is not None),
            "_action_file_name variable must be overwritten by the class that extends JobBase",
        )
        return conf.get_action_file_name(cls._action_file_name_conf_key)



from argparse import ArgumentParser, Namespace
from typing import Dict

from karadoc.common import conf
from karadoc.common.commands.command import Command
from karadoc.common.commands.options.dry_option import DryOption
from karadoc.common.commands.options.read_from_option import ReadFromOption
from karadoc.common.commands.options.tables_option import TablesOption
from karadoc.common.commands.options.vars_option import VarsOption
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.commands.spark import init_job
from karadoc.common.commands.utils import run_job_with_logging
from karadoc.common.exceptions import JobDisabledException
from karadoc.common.model import variables
from karadoc.common.quality.exec import (
    execute_alert,
    execute_metric,
    load_runnable_quality_check,
)
from karadoc.common.quality.quality_check_job import QualityCheckJob
from karadoc.common.validations import fail_if_results
from karadoc.common.validations.job_validations import validate_inputs


def __execute_quality_checks(table: str, job: QualityCheckJob, args, vars: Dict[str, str]):
    job.before_all()
    for alert in job.alerts:
        job.before_each()
        if args.alerts is None or alert.name in args.alerts:
            execute_alert(table, alert, vars, args.output_alert_table)
        job.after_each()
    for metric in job.metrics:
        job.before_each()
        if args.metrics is None or metric.name in args.metrics:
            execute_metric(table, metric, vars, args.output_metric_table)
        job.after_each()
    job.after_all()


def _run_quality_check_for_model(args: Namespace, model_id: str, job_vars: Dict[str, str], **kwargs):
    job = load_runnable_quality_check(model_id, job_vars)
    fail_if_results(validate_inputs(job))
    if job.disable:
        if not args.dry:
            raise JobDisabledException()
        else:
            print("WARN: The job has been disabled, It should not be launched")
    if not args.dry:
        init_job(job=job, raw_args=args.raw_args)
        if args.remote_env is not None:
            job._configure_remote_input(args.remote_env)
        __execute_quality_checks(model_id, job, args, job_vars)
    elif args.dry and not conf.is_dev_env():
        init_job(job=job, raw_args=args.raw_args)


class QualityCheckCommand(Command):
    description = "run QUALITY_CHECK.py scripts for the specified tables"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        TablesOption.add_arguments(parser)
        DryOption.add_arguments(parser)
        ReadFromOption.add_arguments(parser)
        VarsOption.add_arguments(parser)

        parser.add_argument(
            "--output-alert-table",
            dest="output_alert_table",
            default=None,
            required=False,
            type=str,
            metavar="TABLE",
            help="Write the generated alert DataFrames to the external_outputs of the specified table."
            "This table must have a valid POPULATE.py file, with a run() method that will be ignored.",
        )
        parser.add_argument(
            "--output-metric",
            dest="output_metric_table",
            default=None,
            required=False,
            type=str,
            metavar="TABLE",
            help="Write the generated metric DataFrames to the external_outputs of the specified table. "
            "This table must have a valid POPULATE.py file, with a run() method that will be ignored.",
        )
        parser.add_argument(
            "--alerts",
            dest="alerts",
            default=None,
            required=False,
            nargs="+",
            type=str,
            metavar="ALERTS",
            help="Run only the alerts with the specified names",
        )
        parser.add_argument(
            "--metrics",
            dest="metrics",
            default=None,
            required=False,
            nargs="+",
            type=str,
            metavar="METRICS",
            help="Run only the metrics with the specified names",
        )

    @staticmethod
    def do_command(args) -> ReturnCode:
        return_code = ReturnCode.Success

        vars_list = variables.expand_vars(args.vars)

        for model_id in args.tables:
            for job_vars in vars_list:
                return_code &= run_job_with_logging(
                    func=_run_quality_check_for_model,
                    job_type=QualityCheckJob,
                    model_id=model_id,
                    job_vars=job_vars,
                    args=args,
                )

        return return_code



from typing import Callable, Dict, List, Optional, Set, Tuple

import graphviz
import networkx as nx

from karadoc.common import table_utils
from karadoc.common.graph.graph_filter import GraphFilter
from karadoc.common.model.table_index import Populate, Table


def build_graph(table_index: Dict[str, Table]) -> nx.DiGraph:
    """Generate the full dependency graph of the model
    :return: a directed graph of table names
    """
    graph: nx.DiGraph = nx.DiGraph()
    for table in table_index.values():
        graph.add_node(table.full_name)
        if table.populate is not None:
            for source_full_name in table.populate.input_tables:
                graph.add_edge(source_full_name, table.full_name)
    return graph


def __breadth_first_search(  # NOSONAR
    graph: nx.DiGraph,
    upwards: bool,
    max_depth: Optional[int],
    starting_nodes: List[str],
    ignored_edges: List[Tuple[str, str]],
    visit_node: Optional[Callable[[str], None]] = None,
    visit_edge: Optional[Callable[[str, str], None]] = None,
):
    """Explore the given 'graph' in the direction specified by the 'upwards' flag, up to a distance of 'max_depth',
    and call the method 'visit_node' on each visited node and 'visit_edge' on each visited edge.
    A depth of 0 or less means no exploration in that direction.
    A depth equal to None means unbounded exploration in that direction.

    :param graph: The initial graph
    :param upwards: Go upstream instead of downstream if set to true
    :param max_depth: The maximal distance to explore
    :param starting_nodes: The nodes from which the exploration is started
    :param visit_node: (Optional) Function of type (node) -> None called for each visited node
    :param visit_edge: (Optional) Function of type (source, dest) -> None called for each visited edge
    :return: nothing but updates new_graph
    """
    depth = 1
    next_to_visit = set(starting_nodes)
    already_seen = set(starting_nodes)

    def visit_neighboring_edges(_node: str) -> None:
        """Visit all the edges starting from this node (or going to this node if upwards=True)"""
        neighbors = graph.predecessors(_node) if upwards else graph.successors(_node)
        for neighbor in neighbors:
            source, dest = (neighbor, _node) if upwards else (_node, neighbor)
            if (source, dest) not in ignored_edges:
                if visit_edge is not None:
                    visit_edge(source, dest)
                if neighbor not in already_seen:
                    already_seen.add(neighbor)
                    next_to_visit.add(neighbor)

    while len(next_to_visit) > 0 and (max_depth is None or depth <= max_depth):
        to_visit = next_to_visit
        next_to_visit = set()
        while len(to_visit) > 0:
            node = to_visit.pop()
            if visit_node is not None:
                visit_node(node)
            visit_neighboring_edges(node)
        depth += 1


def __build_oriented_neighbor_subgraph(
    graph: nx.DiGraph,
    new_graph: nx.DiGraph,
    upwards: bool,
    max_depth: Optional[int],
    starting_nodes: List[str],
    ignored_edges: List[Tuple[str, str]],
):
    """Explore the given 'graph' in the direction specified by the 'upwards' flag, up to a distance of 'max_depth',
    and adds the explored nodes and edges to the given 'new_graph'.
    A depth of 0 or less means no exploration in that direction.
    A depth equal to None means unbounded exploration in that direction

    :param graph: The initial graph
    :param new_graph: The resulting graph that will be build (not necessarily empty at the start)
    :param upwards: Go upstream instead of downstream if set to true
    :param max_depth: The maximal distance to explore
    :param starting_nodes: The nodes from which the exploration is started
    :return: nothing but updates new_graph
    """
    for node in starting_nodes:
        new_graph.add_node(node)

    def visit_edge(source: str, dest: str) -> None:
        if not new_graph.has_edge(source, dest):
            new_graph.add_edge(source, dest)

    __breadth_first_search(
        graph,
        upwards,
        max_depth,
        starting_nodes,
        ignored_edges,
        visit_edge=visit_edge,
    )


def build_graph_filters(
    nodes: List[str], default_upstream_depth: Optional[int], default_downstream_depth: Optional[int]
):
    def to_graph_filter(s):
        if type(s) == GraphFilter:
            return s
        else:
            return GraphFilter(s, default_upstream_depth, default_downstream_depth)

    return [to_graph_filter(s) for s in nodes]


def get_filtered_subgraph(
    graph: nx.DiGraph, graph_filters: List[GraphFilter], ignored_edges: List[Tuple[str, str]]
) -> nx.DiGraph:
    """Generate a new subgraph of the given 'graph' starting from the specified 'nodes',
        exploring upstream up to a depth of 'upstream_depth' and downstream up to a depth of 'downstream_depth'.
        A depth of 0 or less means no exploration in that direction.
        A depth equal to None means unbounded exploration in that direction

    :param graph:
    :param graph_filters:
    :param ignored_edges:
    :return:
    """
    new_graph: nx.DiGraph = nx.DiGraph()
    for graph_filter in graph_filters:
        __build_oriented_neighbor_subgraph(
            graph, new_graph, True, graph_filter.upstream_depth, [graph_filter.node], ignored_edges
        )
        __build_oriented_neighbor_subgraph(
            graph, new_graph, False, graph_filter.downstream_depth, [graph_filter.node], ignored_edges
        )
    return new_graph


def get_orphans(graph: nx.DiGraph) -> List[str]:
    """Returns the list of nodes without any predecessors.

    :param graph:
    :return:
    """
    res = []
    for node in graph.nodes:
        if next(graph.predecessors(node), 0) == 0:
            res.append(node)
    return res


def find_tables_to_disable(index: Dict[str, Table], graph: nx.DiGraph) -> List[str]:
    """Find tables that can be disabled. A table can be disabled if all three conditions below are met :
    - it has no external output
    - all of its direct successors are already disabled or can be disabled (recursively)

    :param index:
    :param graph:
    :return:
    """
    reverse_topological_sort = reversed(get_topological_sort(graph))
    dead_tables = {}

    def _is_disabled(populate: Populate):
        return populate.disabled

    def _has_external_outputs(populate: Populate):
        return len(populate.external_outputs) > 0

    def _all_direct_successors_are_dead(table):
        for successor in graph.successors(table):
            if successor not in dead_tables:
                return False
        return True

    for table in reverse_topological_sort:
        table_info = index[table]
        populate = table_info.populate
        if populate is not None:
            if _is_disabled(populate):
                dead_tables[table] = "already disabled"
            elif not _has_external_outputs(populate) and _all_direct_successors_are_dead(table):
                dead_tables[table] = "can be disabled"

    return sorted([table for table, state in dead_tables.items() if state == "can be disabled"])


def get_topological_sort(graph: nx.DiGraph) -> List[str]:
    return list(nx.algorithms.topological_sort(graph))


def _format_table_node(table: str, table_name: str, populate: Populate, highlight_tables: Set[str]) -> Dict[str, str]:
    """Build the set of format option that will be passed to graphviz for this table node.

    :param table: full name of the table
    :param table_name: short name of the table
    :param populate: populate representing the table
    :param highlight_tables: set of full table names that should be highlighted
    :return: a Dict of formatting options
    """
    node_format = {
        "shape": "box",
        "fillcolor": "aliceblue",
        "label": table_name,
        "style": "filled",
        "color": None,
    }
    if populate is not None:
        if len(populate.partitions) > 0:
            node_format["label"] += "/%s" % "/".join(populate.partitions)
            node_format["fillcolor"] = "bisque"
        if populate.disabled:
            node_format["fillcolor"] = "gray90"
    if table in highlight_tables:
        node_format["style"] += ",bold"
        node_format["color"] = "blue"
    return node_format


def render_graph(
    graph: nx.DiGraph,
    table_index: Dict[str, Table],
    highlight_tables: Optional[List[str]] = None,
    output_format: str = "png",
) -> graphviz.Digraph:
    """Generates a Digraph.gv file and display it using the Graphviz library"""
    if highlight_tables is None:
        highlight_tables = set()
    else:
        highlight_tables = set(highlight_tables)
    dot = graphviz.Digraph(format=output_format)
    dot.attr(rankdir="LR")
    schema_subgraphs = {}

    def get_schema_subgraph(schema: str) -> graphviz.Digraph:
        sub = schema_subgraphs.get(schema)
        if sub is None:
            sub = graphviz.Digraph(name="cluster_" + schema)
            sub.attr(label=schema)
            sub.attr(style="dotted")
            schema_subgraphs[schema] = sub
        return sub

    for table in graph.nodes():
        table_info = table_index[table]
        (schema_name, table_name, _) = table_utils.parse_table_name(table)
        sub = get_schema_subgraph(schema_name)
        populate: Populate = table_info.populate
        node_format = _format_table_node(table, table_name, populate, highlight_tables)
        sub.node(table, **node_format)
    for (source, dest) in graph.edges():
        dot.edge(source, dest)

    for (schema, sub) in schema_subgraphs.items():
        dot.subgraph(sub)

    dot.render(view=True)

    return dot



from karadoc.common.job_core.has_inputs import HasInputs


class HasStreamInputs(HasInputs):
    def __init__(self) -> None:
        super().__init__()

    def get_reader(self):
        return self.spark.readStream



import os
from functools import lru_cache
from typing import Dict, Generator, Iterable, Optional, Tuple, Type

from karadoc.common import conf
from karadoc.common.job_core.job_base import JobBase
from karadoc.common.utils.assert_utils import assert_true


def _is_hidden(folder: str):
    """We consider that a folder is hidden if at least one of it's ancestor folder's name starts with a '.' or '_'"""
    normalized_path = os.path.normpath(folder)
    ancestors = normalized_path.split(os.sep)
    for ancestor in ancestors:
        if ancestor.startswith(".") or ancestor.startswith("_"):
            return True
    return False


# We use a cache here to avoid recomputing everything all the time.
@lru_cache(maxsize=100)
def __index_files_with_name_cached(file_name) -> Dict[str, Dict[str, str]]:
    """Recursively scan the model folder and returns all the relative paths of the files
    indexed by schema and table names
    :param file_name: name of the files to match
    :return:
    """
    schemas = {}
    for schema, table, folder in list_schema_table_folders():
        for file in os.listdir(folder):
            if file == file_name:
                schemas.setdefault(schema, {})[table] = os.path.join(folder, file)

    return schemas


def __index_files_with_name(file_name) -> Dict[str, Dict[str, str]]:
    """Recursively scan the model folder and returns all the relative paths of the files
    indexed by schema and table names
    :param file_name: name of the files to match
    :return:
    """
    # For unit tests, we disable the cache because the model directory may change between tests.
    if not conf.enable_file_index_cache():
        __index_files_with_name_cached.cache_clear()
    return __index_files_with_name_cached(file_name)


def list_schema_table_folders() -> Generator:
    """Recursively scan the model folder and returns all the triplets
        (schema_name, table_name, table_folder_path)
    :return: A generator with all (schema_name, table_name, table_folder_path)
    """
    model_location = conf.get_model_folder_location()
    for (dir, subdirs, files) in os.walk(model_location):
        if not _is_hidden(os.path.relpath(dir, model_location)) and dir.endswith(".db"):
            for subdir in subdirs:
                if not subdir.startswith(".") and not subdir.startswith("_"):
                    schema_name = os.path.split(dir)[1][:-3]
                    table_name = subdir
                    yield schema_name, table_name, os.path.join(dir, subdir)


def __get_indexed_action_file(action_file_name: str) -> Dict[str, Dict[str, str]]:
    """Recursively scan the model folder and returns all ACTION_FILE.py files relative paths
    indexed by schema and table names
    :return: A dict of dict
    """
    return __index_files_with_name(f"{action_file_name}.py")


def list_action_files(job_type: Type[JobBase]) -> Iterable[Tuple[str, str, str]]:
    """List all action files corresponding to jobs of type job_type in the project
    :param job_type: job type defined in the action file. (e.g. SparkBathJob, QualityCheckJob ...)
    :return: A generator with all STREAM.py paths
    """
    assert_true(issubclass(job_type, JobBase))
    for schema, tables in __get_indexed_action_file(job_type.get_action_file_name()).items():
        for table, stream in tables.items():
            yield schema, table, stream


def get_action_file(schema_name: str, table_name: str, job_type: Type[JobBase]) -> Optional[str]:
    """Recursively scan the model folder and return the path of the ACTION_FILE.py file for the given table_name
    :param schema_name: name of the schema
    :param table_name: name of the table
    :param job_type: job type defined in the action file. (e.g. SparkBathJob, QualityCheckJob ...)
    :return: A string path
    """
    assert_true(issubclass(job_type, JobBase))
    result = __get_indexed_action_file(job_type.get_action_file_name()).get(schema_name, {}).get(table_name)
    if result:
        return result
    else:
        return None



import os
from typing import Dict

from google.cloud import bigquery
from google.cloud.exceptions import NotFound
from google.oauth2 import service_account
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.streaming import DataStreamWriter

from karadoc.common.conf import ConfBox
from karadoc.common.conf.configurable_class import ConfParam
from karadoc.common.connector import ConfigurableConnector
from karadoc.common.table_utils import parse_table_name
from karadoc.common.utils.assert_utils import assert_true


def _get_blob_location(bucket_name: str, table_name: str) -> str:
    (schema_name, table_name, _) = parse_table_name(table_name)
    return "gs://%s/tmp/big-query/%s.db/%s" % (bucket_name, schema_name, table_name)


class BigQueryConnector(ConfigurableConnector):
    project_name = ConfParam(
        required=True, description="Name of the bigquery project where the source/destination table is located."
    )
    parent_project = ConfParam(
        default=lambda conn: conn.project_name.get(), description="Name of the project used for billing."
    )
    bucket_name = ConfParam(
        required=False,
        description="(Required for writing) Name of the intermediary Google Storage bucket into which "
        "Spark will write data before BigQuery ingests them. The used service account "
        "must have write access rights on this bucket.",
    )
    json_keyfile = ConfParam(required=True, description="Path to the json file containing GCP authentication secrets.")
    materialization_dataset = ConfParam(
        description="(Required for reading views or custom queries) "
        "Name of the dataset where temporary BigQuery results will be written. "
        "The used service account must have write access rights on this dataset."
    )
    materialization_project = ConfParam(
        description="(Required for reading views or custom queries) "
        "Name of the project where temporary BigQuery results will be written. "
        "The used service account must have write access rights on this dataset."
    )

    def __init__(self, spark: SparkSession, connection_conf: ConfBox):
        ConfigurableConnector.__init__(self, spark, connection_conf)

    def __get_json_keyfile_path(self):
        json_keyfile_path = self.json_keyfile.get()
        assert_true(os.path.isfile(json_keyfile_path), "File %s not found" % json_keyfile_path)
        return json_keyfile_path

    def __get_read_options(self) -> Dict[str, str]:
        json_keyfile_path = self.__get_json_keyfile_path()
        read_options = {
            "project": self.project_name.get(),
            "parentProject": self.parent_project.get(),
            "credentialsFile": json_keyfile_path,
        }
        if self.materialization_project.is_defined() and self.materialization_dataset.is_defined():
            read_options.update(
                **{
                    "viewsEnabled": "true",
                    "materializationDataset": self.materialization_dataset.get(),
                    "materializationProject": self.materialization_project.get(),
                }
            )

        return read_options

    def __get_write_options(self, dest) -> Dict[str, str]:
        json_keyfile_path = self.__get_json_keyfile_path()
        write_options = {
            "project": self.project_name.get(),
            "parentProject": self.parent_project.get(),
            "intermediateFormat": dest.get("loading_format", "parquet"),
            "temporaryGcsBucket": self.bucket_name.get(required=True),
            "credentialsFile": json_keyfile_path,
        }
        gs_options = {
            "fs.gs.impl": "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFileSystem",
            "fs.AbstractFileSystem.gs.impl": "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFS",
            "google.cloud.auth.service.account.enable": "true",
            "fs.gs.auth.service.account.enable": "true",
            "fs.gs.auth.service.account.json.keyfile": json_keyfile_path,
        }
        for k, v in gs_options.items():
            self.spark._jsc.hadoopConfiguration().set(k, v)
        return write_options

    def __get_bigquery_client(self) -> bigquery.Client:
        credentials = service_account.Credentials.from_service_account_file(self.__get_json_keyfile_path())
        return bigquery.Client(project=self.project_name.get(), credentials=credentials)

    @staticmethod
    def __is_an_existing_table(client: bigquery.Client, table_name: str) -> bool:
        try:
            client.get_table(table_name)
            return True
        except NotFound:
            return False

    def write(self, df: DataFrame, dest):
        """Write the given DataFrame to an external output on Google BigQuery via the
        `spark-bigquery connector <https://github.com/GoogleCloudPlatform/spark-bigquery-connector>`_

        The destination must have the following attributes:

        - "table": full name of the destination table (e.g. "dataset_name.table_name")
        - "mode": Loading mode used by spark:

          - overwrite: erase previously existing data (recommended because idempotent)
          - append: append to previously existing data (useful for incremental insert but not idempotent)
          - errorifexists: fail if there is previously existing data (not idempotent)
          - ignore: do nothing if there is previously existing data (not idempotent)
        - "loading_format": (Optional) Format used to write the data on google storage before loading into BigQuery:

          - parquet: (default) Arrays can be null but will be wrapped with ".list.element" fields
          - orc: Null arrays will be converted to empty arrays (less naming boilerplate but you won't be able to tell
            the difference between empty and missing arrays)
          - avro: No extra ".list.element" boilerplate like parquet, but timestamps are converted to Integers.
        - "expiring_date": (Unsupported atm) Datetime in UTC at which the table will be deleted
        - "options": (Optional[Dict[str, str]]) options that can be passed to the bigquery reader as a Dict[str, str]
          some examples `here <https://github.com/GoogleCloudPlatform/spark-bigquery-connector>`_.

        The corresponding connection must have the following attributes:

        - type = 'bigquery'
        - bucket_name = name of the Google Storage bucket to use to write the intermediate temporary file.
        - project_name = name of the BigQuery project where the destination table is located
        - json_keyfile = path to the json file containing the BigQuery credentials

        :param df: the DataFrame to export
        :param dest: A dictionary describing the destination
        :return:
        """
        write_options = self.__get_write_options(dest)
        df.write.format("bigquery").mode(dest["mode"]).options(**write_options).options(**dest.get("options", {})).save(
            dest["table"]
        )

    def write_stream(self, stream_df: DataFrame, dest) -> DataStreamWriter:
        """Write the given stream DataFrame to an external output on Google BigQuery in streaming via the
        `spark-bigquery connector <https://github.com/GoogleCloudPlatform/spark-bigquery-connector>`_

        The destination must have the following attributes:

        - "table": full name of the destination table (e.g. "dataset_name.table_name")
        - "loading_format": (Optional) Format used to write the data on google storage before loading into BigQuery:

          - parquet: (default) Arrays can be null but will be wrapped with ".list.element" fields
          - orc: Null arrays will be converted to empty arrays (less naming boilerplate but you won't be able to tell
            the difference between empty and missing arrays)
          - avro: No extra ".list.element" boilerplate like parquet, but timestamps are converted to Integers.
        - "expiring_date": (Unsupported atm) Datetime in UTC at which the table will be deleted
        - "options": (Optional[Dict[str, str]]) options that can be passed to the bigquery reader as a Dict[str, str]
          some examples `here <https://github.com/GoogleCloudPlatform/spark-bigquery-connector>`_.

        The corresponding connection must have the following attributes:

        - type = 'bigquery'
        - project_name = name of the BigQuery project where the destination table is located
        - bucket_name = name of the Google Storage bucket to use to write the intermediate temporary file.
        - json_keyfile = path to the json file containing the BigQuery credentials

        :param stream_df: the DataFrame to export
        :param dest: A dictionary describing the destination
        :return:
        """
        write_options = self.__get_write_options(dest)
        return (
            stream_df.writeStream.format("bigquery")
            .options(**write_options)
            .options(**dest.get("options", {}))
            .option("table", dest["table"])
        )

    def read(self, source: dict) -> DataFrame:
        """Read data from a Big Query table via the
        `spark-bigquery connector <https://github.com/GoogleCloudPlatform/spark-bigquery-connector>`_

        The source must have the following attributes:

        - "connection": name of the bigquery connection
        - "table": name to the table to be read
        - "query": (Optional) execute the provided query on BigQuery and returns the result.
          if provided, "table" will be ignored but we recommend to still fill it for documentation purposes.
        - "options": (Optional[Dict[str, str]]) options that can be passed to the bigquery reader as a Dict[str, str]
          some examples `here <https://github.com/GoogleCloudPlatform/spark-bigquery-connector>`_.

        The corresponding connection must have the following attributes:

        - type = 'bigquery'
        - project_name = name of the bigquery project where the source table is located
        - bucket_name = name of the bigquery bucket (It is only used for write method)
        - json_keyfile = file of bigquery connection configuration

        :param source: a dictionary describing the source
        :return: DataFrame
        """
        read_options = self.__get_read_options()
        spark_reader = self.spark.read.format("bigquery").options(**read_options)
        spark_reader = spark_reader.options(**source.get("options", {}))

        if source.get("query") is not None:
            self.materialization_dataset.require()
            self.materialization_project.require()
            spark_reader = spark_reader.option("query", source["query"])
        else:
            spark_reader = spark_reader.option("table", source["table"])

        df = spark_reader.load()
        return df

    def remove_partition(self, partition: str, dest):
        """
        The destination must have the following attributes:

        - "connection": name of the bigquery connection
        - "table": name to the table to be read
        - "options": (Optional[Dict[str, str]]) options that can be passed to the bigquery reader as a Dict[str, str]
          some examples `here <https://github.com/GoogleCloudPlatform/spark-bigquery-connector>`_.

        The corresponding connection must have the following attributes:

        - type = 'bigquery'
        - project_name = name of the bigquery project where the source table is located
        - bucket_name = name of the bigquery bucket (It is only used for write method)
        - json_keyfile = file of bigquery connection configuration

        Documentation to delete a partition
        `here <https://cloud.google.com/bigquery/docs/managing-partitioned-tables#delete_a_partition>`_.

        :param partition: The BigQuery partition to remove, express as string, e.g for a DAY: YYYYMMDD
        :param dest: A dictionary describing the destination
        :return:
        """

        bq = self.__get_bigquery_client()
        if self.__is_an_existing_table(bq, dest["table"]):
            bq.delete_table("{table}${partition}".format(table=dest["table"], partition=partition))
        bq.close()



from argparse import ArgumentParser, Namespace

from karadoc.common.commands.command import Command
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.model.connection_index import list_connections_df
from karadoc.common.output.local_export import local_export_dataframe


class ListConnectionsCommand(Command):
    description = "list all connections"

    @staticmethod
    def add_arguments(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--env", dest="env", metavar="ENV", type=str, help="List the connections on the specified environment"
        )
        parser.add_argument(
            "--format",
            dest="format",
            default="markdown",
            type=str,
            choices=["markdown", "csv", "xlsx"],
            help="(Default: markdown) Output format for the command results.",
        )
        parser.add_argument(
            "--output",
            dest="output",
            type=str,
            default=None,
            metavar="PATH",
            help="Path of the file where the results of the command will be written. "
            "By default, the results are written on stdout, except for binary format (e.g. xlsx) "
            "for which the default is connections.FORMAT",
        )

    @staticmethod
    def do_command(args: Namespace) -> ReturnCode:
        binary_formats = ["xlsx"]
        if args.output is None and args.format in binary_formats:
            args.output = f"connections.{args.format}"

        df = list_connections_df(args.env)
        local_export_dataframe(df=df, output=args.output, format=args.format)
        return ReturnCode.Success



from typing import Iterable, List, Optional

import pandas as pd
from bigquery_frame import BigQueryBuilder, DataFrame
from bigquery_frame import functions as f
from bigquery_frame.transformations_impl.analyze import analyze
from google.cloud import bigquery
from google.oauth2 import service_account

from karadoc.common import conf
from karadoc.common.output.local_export import write_pandas_dataframes_to_xlsx
from karadoc.common.utils.assert_utils import assert_true


def _generate_empty_line_df(bq: BigQueryBuilder) -> DataFrame:
    df = bq.sql(
        """
        SELECT
            CAST(NULL as STRING) as `table_name`,
            CAST(NULL as STRING) as `column_name`,
            CAST(NULL as STRING) as `column_type`,
            CAST(NULL as INT) as `count`,
            CAST(NULL as INT) as `count_distinct`,
            CAST(NULL as INT) as `count_null`,
            CAST(NULL as STRING) as `min`,
            CAST(NULL as STRING) as `max`,
            CAST(NULL as STRING) as `approx_top_100`,
        """
    )
    return df


def _get_gcp_credentials_path(env: str):
    env_conf = conf.get_connection_conf(env)
    assert_true("bigquery" in env_conf.get("type"))
    json_keyfile = env_conf.get("json_keyfile")
    return json_keyfile


def _get_bq_client(gcp_credentials_path: str) -> bigquery.Client:
    if gcp_credentials_path.endswith(".json"):
        credentials = service_account.Credentials.from_service_account_file(
            gcp_credentials_path,
            scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )
        client = bigquery.Client(credentials=credentials, project=credentials.project_id)
        return client
    else:
        raise ValueError(f"Expected a .json file, got : {gcp_credentials_path}")


def _analyze_table(table_name: str, bq: BigQueryBuilder) -> Iterable[DataFrame]:
    print(table_name + ":")
    df = analyze(bq.table(table_name)).persist()
    MAX_CHARS = 20
    df = df.withColumn(
        "approx_top_100",
        f.expr(
            f"""
            ARRAY(
              SELECT
                CONCAT(
                    '(',
                    SUBSTRING(x.value, 0, {MAX_CHARS}),
                    CASE WHEN LENGTH(x.value) > 10 THEN '...' ELSE '' END,
                    ',',
                    CAST(x.count as STRING),
                    ')'
                )
              FROM UNNEST(approx_top_100) x
            )
            """
        ),
        replace=True,
    )
    df = df.withColumn(
        "approx_top_100", f.expr("""CONCAT('[', ARRAY_TO_STRING(approx_top_100, ", ") , ']')"""), replace=True
    )
    yield df.select(f.lit(table_name).alias("table_name"), *df.columns)
    yield _generate_empty_line_df(bq)


def analyze_tables(tables: List[str], env: str, output: Optional[str]):
    json_keyfile = _get_gcp_credentials_path(env)
    client = _get_bq_client(json_keyfile)
    bq = BigQueryBuilder(client)

    pdfs = [df.toPandas() for table in tables for df in _analyze_table(table, bq)]
    if output is None:
        output = "report.xlsx"

    # We perform the union in Pandas because unions with bigquery-frame do not preserve ordering.
    union_pdf = pd.concat(pdfs)
    write_pandas_dataframes_to_xlsx(union_pdf, output, index=False, freeze_panes=(1, 0))



import inspect
from typing import Any, Type

from karadoc.common.exceptions import ConnectorLoadError


def _is_defined_here(obj: Any, module) -> bool:
    """Returns true if the given class was defined in the given module.
    This discards classes that are imported in the module but defined elsewhere."""
    return inspect.getfile(obj) == inspect.getfile(module)


def find_class_from_module(module, class_type: Type[Any]):
    """Locates in the file the class defined that extends the specified class and returns it.
    Raises an exception if more than one such class is found in the module.

    :return:
    """
    class_full_name = class_type.__module__ + "." + class_type.__qualname__
    found_classes = [
        obj
        for name, obj in inspect.getmembers(module)
        if inspect.isclass(obj) and issubclass(obj, class_type) and _is_defined_here(obj, module)
    ]
    if len(found_classes) == 1:
        return found_classes[0]
    if len(found_classes) == 0:
        raise ConnectorLoadError(
            f"Error while loading the module named '{module.__name__}' : "
            f"no class extending {class_full_name} was found"
        )
    if len(found_classes) > 1:
        raise ConnectorLoadError(
            f"Error while loading the module named '{module.__name__}' : "
            f"we found more than one class extending {class_full_name}\n"
            f"Classes found: {found_classes}"
        )



from pyspark.sql import DataFrame

from karadoc.common.job_core.has_output import HasOutput
from karadoc.spark.conf import (
    get_batch_default_output_format,
    get_write_options_for_format,
)


class HasBatchOutput(HasOutput):
    def __init__(self) -> None:
        super().__init__()
        self.output_format = get_batch_default_output_format()
        self.output_mode = "OVERWRITE"

    def write_table(self, df: DataFrame):
        self._write_table(
            df=df,
            path=self.hdfs_output(),
            output_format=self.output_format,
            output_mode=self.output_mode,
            output_options=self.output_options,
            partitions=self.dynamic_partitions(),
        )

    @staticmethod
    def _write_table(
        df: DataFrame, path: str, output_format: str, output_mode: str, output_options=None, partitions=None
    ):
        if output_options is None:
            output_options = {}
        if partitions is None:
            partitions = []
        writer = df.write.partitionBy(partitions).mode(output_mode).format(output_format)
        write_options = get_write_options_for_format(output_format)
        options = {**write_options, **output_options}
        writer.options(**options).save(path)



from typing import Dict, Optional

from karadoc.common.conf.conf_box import ConfBox
from karadoc.common.conf.package import _get_settings_for_env

SPARK_CONF_GROUP = "spark.conf"


def get_spark_conf(env: Optional[str] = None) -> Dict[str, object]:
    """Return the configuration for the given connection in the specified environment.
    If no environment is specified, use the current environment.

    :param env:
    :return:
    """
    dynabox = _get_settings_for_env(env).get(SPARK_CONF_GROUP)
    if dynabox is None:
        return {}
    else:
        return ConfBox(dynabox).to_flat_dict()


def get_batch_default_output_format(env: Optional[str] = None) -> str:
    return (
        _get_settings_for_env(env)
        .get("spark", {})
        .get("write", {})
        .get("batch", {})
        .get("default_format", default="parquet")
    )


def get_stream_default_output_format(env: Optional[str] = None) -> str:
    return (
        _get_settings_for_env(env)
        .get("spark", {})
        .get("write", {})
        .get("stream", {})
        .get("default_format", default="parquet")
    )


def get_write_options_for_format(output_format: str, env: Optional[str] = None) -> Dict:
    """Return the configured write options for the given format

    Example:

    In your settings.toml:

    .. code-block:: python

        [default.spark.write.options.text]
            compression = "gzip"

        or

        [default.spark.write.options]
          text.compression = "gzip"

    In your code:

    compression = get_write_options_for_format("text")
    """
    format_write_options = (
        _get_settings_for_env(env).get("spark", {}).get("write", {}).get("options", {}).get(output_format)
    )
    if format_write_options is not None:
        return format_write_options.to_dict()
    else:
        return {}



import textwrap
from argparse import HelpFormatter
from typing import List


class LineBreakHelpFormatter(HelpFormatter):
    """Help message formatter which preserves line breaks descriptions."""

    def _split_lines(self, text: str, width: int) -> List[str]:
        res = []
        for lines in text.split("\n"):
            res += textwrap.wrap(lines, width)
        return res



from distutils.util import strtobool
from typing import Dict, Optional

from karadoc.common import conf
from karadoc.common.exceptions import MissingVariablesException
from karadoc.common.utils.assert_utils import assert_true


class HasVars:
    """When vars are passed via the --vars option in the command line,
    we have to set them globally to let the Job object override it's own variables with it.
    """

    _global_vars: Optional[Dict[str, str]] = None
    __supported_types = [str, bool, int, float]
    __supported_types_names = [__cls.__name__ for __cls in __supported_types]

    def __init__(self) -> None:
        # Attributes that the user is not supposed to change
        self.__vars_override = HasVars._global_vars
        self.__vars: Dict[str, str] = dict()

    @property
    def vars(self) -> Dict[str, str]:
        """Variables that can be used inside the job and can be overwritten with the '--vars' option"""
        return self.__vars

    @vars.setter
    def vars(self, new_vars: Dict[str, str]):
        """Setting vars in a ACTION_FILE behave differently depending on the context.

        - When a job object is created directly in an interactive shell or in a ACTION_FILE loaded
          with the `load_non_runnable_action_file` methods, `global_vars` is None and the vars are set normally.

        - When a job object is created in an ACTION_FILE loaded with the `load_runnable_action_file` methods
          `global_vars` is not None and the vars are overridden by them. When `job.vars = {...}` is called
          inside the ACTION_FILE, we will not set them but only check that the global_vars correspond to the
          variables declared in the ACTION_FILE.

            When we run a job with declared variables (job.vars), the values of the declared variables
          must be passed via the --vars option.
          This security can be disabled on an environment by setting allow_missing_vars = true (default: false),
          in which case the default values will be used when missing, instead of raising an error.
          It can be useful for development environment, but it is recommended to keep this to false in production.

        :return:
        """
        self.__check_vars_type(new_vars)
        if self.__vars_override is None:
            self.__vars = new_vars
        else:
            if not conf.allow_missing_vars():
                self.__check_vars(new_vars, self.__vars_override)
            self.__vars = {k: self.__change_var_type(self.__vars_override.get(k, v), v) for k, v in new_vars.items()}

    @staticmethod
    def __change_var_type(value: str, template_value: object):
        """Cast the given value to the same type of the template_value.

        Only the following types are supported for now:
        - str
        - bool
        - int
        - float

        """
        if isinstance(template_value, str):
            return value
        elif isinstance(template_value, bool):
            assert_true(
                isinstance(value, str) or isinstance(value, bool),
                f"Expected str or bool, got {type(value).__name__} instead.",
            )
            if isinstance(value, str):
                return strtobool(value)
            else:
                return value
        elif isinstance(template_value, int):
            return int(value)
        elif isinstance(template_value, float):
            return float(value)
        else:
            raise TypeError(
                f"Type {type(template_value).__name__} is not supported for variables. "
                f"Only the following types are supported ${HasVars.__supported_types_names}"
            )

    @staticmethod
    def __check_vars(declared_vars: dict, provided_vars: dict):
        """Check that the all the declared variables are declared in the provided variables"""
        missing_keys = []
        for key in declared_vars.keys():
            if not provided_vars.__contains__(key):
                missing_keys.append(key)
        if len(missing_keys) > 0:
            raise MissingVariablesException(missing_keys=missing_keys)

    @staticmethod
    def __is_var_type_supported(value: object) -> bool:
        """Return true if the type of the given variable is supported"""
        for tpe in HasVars.__supported_types:
            if isinstance(value, tpe):
                return True
        return False

    @staticmethod
    def __check_vars_type(vars: Dict[str, object]):
        """Check that the type of the given variables is supported"""
        for key, value in vars.items():
            if not isinstance(key, str):
                raise ValueError(
                    f"Only strings are allowed as vars keys. Got key: {key} of type: {type(key).__name__} instead."
                )
            if not HasVars.__is_var_type_supported(value):
                raise ValueError(
                    f"Type {type(value).__name__} of variable '{key}' is not supported for variables. "
                    f"Only the following types are supported {HasVars.__supported_types_names}"
                )



from abc import ABC
from argparse import ArgumentParser

from karadoc.common.commands.command import Command


class VarsOption(Command, ABC):
    @staticmethod
    def add_arguments(parser: ArgumentParser) -> None:
        from karadoc.common.commands.parsing import StoreDictKeyValue

        parser.add_argument(
            "--vars",
            nargs="*",
            dest="vars",
            metavar="KEY=VALUE",
            required=False,
            action=StoreDictKeyValue,
            help="Set values for job variables",
        )



import inspect
from typing import Callable, Dict, Iterable, Type

from karadoc.common.conf import ConfBox
from karadoc.common.exceptions import MissingConfigurationParameterException
from karadoc.common.validations import (
    ValidationResult,
    ValidationResultTemplate,
    ValidationSeverity,
)

ValidationResult_SecretParam = ValidationResultTemplate(
    check_type="karadoc.conf.secret_param",
    message_template="Parameter {param_name} should be secret",
    default_severity=ValidationSeverity.Warning,
)

ValidationResult_WrongParamType = ValidationResultTemplate(
    check_type="karadoc.conf.param_type",
    message_template="Incorrect parameter type for {param_name}, "
    "expected {expected_type} "
    "and got {actual_type} instead",
    default_severity=ValidationSeverity.Error,
)

ValidationResult_MissingRequiredParam = ValidationResultTemplate(
    check_type="karadoc.conf.missing_required_param",
    message_template="Missing required parameter {param_name}",
    default_severity=ValidationSeverity.Error,
)

ValidationResult_UnknownParam = ValidationResultTemplate(
    check_type="karadoc.conf.unknown_param",
    message_template="Unknown parameter {param_name}",
    default_severity=ValidationSeverity.Error,
)


class ConfParam:
    def __init__(
        self,
        required: bool = False,
        default: object = None,
        description: str = None,
        type: Type = None,
        secret: bool = False,
    ):
        """Declares a parameter of a `ConfigurableClass`

        :param required: indicates if the parameter is required (see method `ConfigurableClass.validate_params`)
        :param default: default value of the parameter when it is missing (ignored if required=True)
        :param description: description of the parameter
        :param type: type of the parameter (see method `ConfigurableClass.validate_params`)
        :param secret: indicates if the parameter should be stored in a secret store
        """
        self.required = required
        self.default = default
        self.description = description
        self.name = None
        self.type = type
        self.secret = secret
        self._parent = None

    def __str__(self) -> str:
        return (
            f"ConfParam(name={self.name}, type={self.type.__name__}, "
            f"required={self.required}, default={self.default}, secret={self.secret})"
        )

    def get(self, required: bool = None, default=None):
        """Returns the value of this parameter, fetched from the conf.

        Arguments may be passed to this method to override locally the settings of this parameter

        :param required: Locally overrides the defined value of "required" for this parameter
        :param default: Locally overrides the defined value of "default" for this parameter
        :return:
        """
        if default is not None:
            returned_default = default
        elif isinstance(self.default, ConfParam):
            returned_default = self.default.get()
        elif isinstance(self.default, Callable):
            returned_default = self.default(self._parent)
        else:
            returned_default = self.default
        if required is None:
            required = self.required
        if required:
            self.require()
        return self._parent.conf.get(self.name, returned_default)

    def require(self) -> None:
        """Raises an exception if this param is not specified in the configuration"""
        if self.name not in self._parent.conf:
            raise MissingConfigurationParameterException(param_name=self.name)

    def is_defined(self) -> bool:
        return self.name in self._parent.conf


class ConfigurableClass:
    """Abstract class that other classes may inherit to declare and use ConfParams.

    **Examples**
    *(more examples available in the documentation of the `validate_params` method)*

    >>> class MyClass(ConfigurableClass):  # noqa: E501
    ...    a = ConfParam()                                        # A: Parameter values may be passed via at class instantiation via a dict-like conf
    ...    b = ConfParam()                                        # B: Missing parameters that are not required will be None
    ...    c = ConfParam(default="c")                             # C: Parameters may have a default value
    ...    d = ConfParam(default=a)                               # D: The default value can be another parameter
    ...    e = ConfParam(default=lambda x: x.a.get() + x.c.get()) # E: The default value can be the result of a combination of other parameters
    ...
    ...    def __init__(self, connection_conf):
    ...        super().__init__(connection_conf)
    >>> my_class = MyClass({'a': 'a'})
    >>> list(my_class.validate_params()) # Call this for immediate validation
    []
    >>> my_class.a.get() # A: Parameter values may be passed via at class instantiation via a dict-like conf
    'a'
    >>> my_class.b.get() is None # B: Missing parameters that are not required will be None
    True
    >>> my_class.c.get() # C: Parameters may have a default value
    'c'
    >>> my_class.c.get(default='z') # C: The default value may be locally overridden
    'z'
    >>> my_class.c.get(required=True) # C: Param may be made required locally
    Traceback (most recent call last):
      ...
    karadoc.common.exceptions.package.MissingConfigurationParameterException: Missing required parameter c
    >>> my_class.d.get() # D: The default value can be another parameter
    'a'
    >>> my_class.e.get() # E: The default value can be the result of a combination of other parameters
    'ac'
    """

    def __init__(self, conf: Dict):
        """
        :param conf: a nested dict-like object (a dict of dict, a parsed json, a dynaconf's DynaBox, etc.)
        """
        self.conf = conf
        self.params: Dict[str, ConfParam] = {
            name: param for name, param in inspect.getmembers(self) if isinstance(param, ConfParam)
        }
        for name, param in self.params.items():
            param.name = name
            param._parent = self

    def validate_params(self) -> Iterable[ValidationResult]:
        """Validates the correctness of all parameters passed to the inheriting class.
        This method may be called immediately in the `__init__()` call of the inheriting class for immediate validation,
        our later for lazy evaluation, depending on the use cases.

        Validation performed:
        - Required params must be present in the passed configuration (even if they have a default value)
        - The type of each param in the passed configuration must match the type in the declaration, when it is declared

        >>> class MyClass(ConfigurableClass):
        ...    a = ConfParam(type=int)                  # A: Parameters should have the correct declared type
        ...    b = ConfParam(required=True)             # B: required parameters must have a value
        ...    c = ConfParam(required=True, default=3)  # C: default values are useless for required parameters
        ...
        ...    def __init__(self, connection_conf):
        ...        super().__init__(connection_conf)
        >>> my_class_1 = MyClass({'b': 2, 'c': 2})
        >>> my_class_1.conf
        {'b': 2, 'c': 2}
        >>> list(my_class_1.validate_params())
        []
        >>> my_class_2 = MyClass({'a': 'a', 'z': 'z'}) # E: passed variables should have the correct declared type
        >>> for res in my_class_2.validate_params(): print(res)
        Error: karadoc.conf.param_type: Incorrect parameter type for a, expected int and got str instead
        Error: karadoc.conf.missing_required_param: Missing required parameter b
        Error: karadoc.conf.missing_required_param: Missing required parameter c
        Error: karadoc.conf.unknown_param: Unknown parameter z

        """
        for name, param in self.params.items():
            yield from _check_secret_params(name, self.conf, param)
            yield from _check_param_types(name, self.conf, param)
            yield from _check_required_param(name, self.conf, param)
        for name in self.conf.keys():
            yield from _check_unknown_param(name, self.params)


def __is_secret_conf(conf: dict, name: str):
    return isinstance(conf, ConfBox) and conf.is_secret(name)


def _check_secret_params(name: str, conf: Dict, param: ConfParam) -> Iterable[ValidationResult]:
    if name in conf and param.secret and not __is_secret_conf(conf, name):
        yield ValidationResult_SecretParam(param_name=name)


def _check_param_types(name: str, conf: Dict, param: ConfParam) -> Iterable[ValidationResult]:
    if name in conf and param.type is not None and not __is_secret_conf(conf, name):
        value = conf.get(name)
        if not isinstance(value, param.type):
            yield ValidationResult_WrongParamType(
                param_name=name, expected_type=param.type.__name__, actual_type=type(value).__name__
            )


def _check_required_param(name: str, conf: Dict, param: ConfParam) -> Iterable[ValidationResult]:
    if name not in conf and param.required:
        yield ValidationResult_MissingRequiredParam(param_name=name)


def _check_unknown_param(name, params: Dict[str, ConfParam]):
    if name not in params:
        yield ValidationResult_UnknownParam(param_name=name)



import sys
from contextlib import contextmanager
from io import StringIO
from typing import Generator, Tuple


@contextmanager
def captured_output() -> Generator[Tuple[StringIO, StringIO], None, None]:
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err



from typing import Dict, Union

from pyspark.sql import DataFrame

from karadoc.common.conf import CONNECTION_GROUP
from karadoc.common.job_core.has_spark import HasSpark


def _write_external_output_signature_check(df: DataFrame, dest: Dict) -> None:
    """We uses this empty method when to check the signature of the equivalent method defined in the POPULATE files"""
    pass


def _write_external_outputs_signature_check(df: DataFrame) -> None:
    """We uses this empty method when to check the signature of the equivalent method defined in the POPULATE files"""
    pass


class HasExternalOutputs(HasSpark):
    def __init__(self) -> None:
        super().__init__()
        # Attributes that the user may change
        self.external_outputs: Dict[str, dict] = {}

        # Attributes that the user is not supposed to change
        self.__write_external_output = None
        self.__write_external_outputs = None

    def _write_external_output_default(self, df: DataFrame, dest: Dict) -> None:
        connector = self.get_output_connector(dest)
        connector.write(df, dest)

    def _write_external_outputs_default(self, df: DataFrame) -> None:
        for _, dest in self.external_outputs.items():
            self.write_external_output(df, dest)

    def write_external_output(self, df: DataFrame, dest: Dict) -> None:
        """Writes a given DataFrame to a given external output

        :param df: The DataFrame to write
        :param dest: The alias of the external output to write to
        :return: nothing
        """
        if self.__write_external_output is None:
            return self._write_external_output_default(df, dest)
        else:
            return self.__write_external_output(df, dest)

    def write_external_outputs(self, df: DataFrame) -> None:
        """Writes a given DataFrame to all the declared external outputs

        :param df: The DataFrame to write
        :return: nothing
        """
        if self.__write_external_outputs is None:
            return self._write_external_outputs_default(df)
        else:
            return self.__write_external_outputs(df)

    def get_output_connector(self, dest: Union[str, Dict]):
        if type(dest) == str:
            dest = self.external_outputs[dest]
        from karadoc.common.connector import load_connector

        return load_connector(dest[CONNECTION_GROUP], self.spark)



from pprint import pprint  # noqa: F401 (used only in doctest)
from typing import Any, Dict, Iterable


def pretty_format(obj: object, indent: int = 2) -> str:
    """Pretty prints an object serialized as json.
    We don't use the pprint method because the output it does not display Dict[Dict[str]] as we want:

    Examples: It must look prettier than pprint
    >>> my_dict={'a':1, 's':{'b': 2, 'c': [3, 4]}}
    >>> pprint(my_dict, indent=2, width=1)
    { 'a': 1,
      's': { 'b': 2,
             'c': [ 3,
                    4]}}
    >>> print(pretty_format(my_dict))
    {
      "a": 1,
      "s": {
        "b": 2,
        "c": [3, 4]
      }
    }

    It must split long lists:
    >>> my_dict={'a':1, 's':{'b': 2, 'c': [33333333333333333333333333333, 44444444444444444444444444444444]}}
    >>> print(pretty_format(my_dict))
    {
      "a": 1,
      "s": {
        "b": 2,
        "c": [
          33333333333333333333333333333,
          44444444444444444444444444444444
        ]
      }
    }

    It must support non-serializable objects:
    >>> from datetime import datetime
    >>> my_dict={'a':1, 's':{'b': 2, 'c': datetime.strptime("2021-10-28", '%Y-%m-%d')}}
    >>> print(pretty_format(my_dict))
    {
      "a": 1,
      "s": {
        "b": 2,
        "c": 2021-10-28 00:00:00
      }
    }

    :param obj:
    :param indent:
    :return:
    """

    def format_list(iterable: Iterable[Any], prefix: str) -> str:
        new_prefix = prefix + " " * indent
        items = [format_obj(item) for item in iterable]
        res = "[" + (", ").join(items) + "]"
        if len(res) > 50:
            res = "[\n" + new_prefix + (",\n" + new_prefix).join(items) + "\n" + prefix + "]"
        return res

    def format_dict(d: Dict[Any, Any], prefix: str) -> str:
        new_prefix = prefix + " " * indent
        items = [format_obj(k, new_prefix) + ": " + format_obj(v, new_prefix) for k, v in d.items()]
        return "{\n" + new_prefix + (",\n" + new_prefix).join(items) + "\n" + prefix + "}"

    def format_obj(o: Any, prefix: str = "") -> str:
        if isinstance(o, str):
            return '"' + o + '"'
        if isinstance(o, Dict):
            return format_dict(o, prefix)
        if isinstance(o, Iterable):
            return format_list(o, prefix)
        return str(o)

    return format_obj(obj)



import logging
import sys
from typing import Optional

from dynaconf.utils.boxing import DynaBox

from karadoc.common.conf.package import get_observability_conf
from karadoc.common.observability import LogEvent
from karadoc.common.observability.logging import setup_logging

LOG = logging.getLogger(__name__)
OBSERVABILITY_CONF_GROUP = "observability"

already_setup = False
observability: Optional[DynaBox] = None


def setup_observability(reset: bool = False, verbose: Optional[bool] = None) -> None:
    """Configure observability on the current process"""
    global already_setup
    if already_setup and not reset:
        return
    observability_conf = get_observability_conf()
    setup_logging(observability_conf, verbose)
    sys.excepthook = __exception_hook
    already_setup = True


def __exception_hook(_exc_type, value, _exc_tb) -> None:
    """Executed on uncaught exceptions"""
    LOG.error(LogEvent("Some uncaught crash occurred", event_type="uncaught_crash"), exc_info=value)



import json

from pyspark.sql import SparkSession

from karadoc.common.conf import ConfBox
from karadoc.common.conf.configurable_class import ConfParam
from karadoc.common.connector import ConfigurableConnector
from karadoc.common.utils.assert_utils import assert_true


class EventHubConnector(ConfigurableConnector):
    HubName = ConfParam(required=True, type=str, description="Name of the Event Hub.")
    SharedAccessKeyName = ConfParam(
        required=True, type=str, description="Name of the shared access policy (SAS Policy)"
    )
    SharedAccessKey = ConfParam(
        required=True,
        type=str,
        secret=True,
        description="Key of the shared access policy",
    )

    def __init__(self, spark: SparkSession, connection_conf: ConfBox):
        ConfigurableConnector.__init__(self, spark, connection_conf)

    def _configure(
        self,
        entity_path,
        starting_time,
        ending_time,
        consumer_group,
        offset=None,
        max_events_per_trigger=None,
        options=None,
    ):
        if options is None:
            options = {}

        hub_name = self.HubName.get()
        shared_access_key_name = self.SharedAccessKeyName.get()
        shared_access_key = self.SharedAccessKey.get()

        connection_string = (
            f"Endpoint=amqps://{hub_name}.servicebus.windows.net/{hub_name};EntityPath={entity_path};"
            f"SharedAccessKeyName={shared_access_key_name};SharedAccessKey={shared_access_key}"
        )

        starting_event_position = {
            "offset": offset,  # not in use
            "seqNo": -1,  # not in use
            "enqueuedTime": starting_time,
            "isInclusive": True,
        }

        ending_event_position = {
            "offset": None,  # not in use
            "seqNo": -1,  # not in use
            "enqueuedTime": ending_time,
            "isInclusive": False,
        }
        max_events_per_trigger_conf = {}
        if max_events_per_trigger is not None:
            max_events_per_trigger_conf = {"maxEventsPerTrigger": max_events_per_trigger}
        eh_conf = {
            "eventhubs.connectionString": self.spark._jvm.org.apache.spark.eventhubs.EventHubsUtils.encrypt(
                connection_string
            ),
            "eventhubs.startingPosition": json.dumps(starting_event_position),
            "eventhubs.endingPosition": json.dumps(ending_event_position),
            "eventhubs.consumerGroup": consumer_group,
            **max_events_per_trigger_conf,
            **options,
        }

        return eh_conf

    @staticmethod
    def _bad_time_format_message(time_var: str) -> str:
        return f"Bad time format for eventhub, should end with 000000Z but got: {time_var}"

    def read(self, source: dict):
        """Read data from an external input via an Event Hub connection.

        The source must have the following attributes:

        - "connection": name of the event_hub connection
        - "entity_path": name of the topic to follow
        - "starting_time": must use the format '%Y-%m-%dT%H:%M:%S.%fZ'
        - "ending_time": must use the format '%Y-%m-%dT%H:%M:%S.%fZ'
        - "consumer_group" (Optional): event_hub consumer group name
        - "options" (Optional): additional options can be added, here is the list of event-hubs-configuration options :
            https://github.com/Azure/azure-event-hubs-spark/blob/master/docs/structured-streaming-eventhubs-integration.md#eventhubsconf

        Because of this bug https://github.com/Azure/azure-event-hubs-spark/issues/444
        we make an assert to prevent retrieving events outside the time bound

        EventHub consumer group documentation:
        https://docs.microsoft.com/en-us/azure/event-hubs/event-hubs-features#event-consumers
        Spark EventHub Connector Documentation:
        https://github.com/Azure/azure-event-hubs-spark/tree/master/docs

        The corresponding event_hub connection must have the following attributes:

        - type = 'event_hub'
        - HubName = name of the hub
        - SharedAccessKeyName = name of the shared access policy (SAS Policy)
        - SharedAccessKey = key of the shared access policy

        :param source: a dictionary describing the source
        :return: a DataFrame (conversationId, destinationAddress, host, message, messageId, messageType, sourceAddress)
        """
        assert_true(
            source["starting_time"][-7:] == "000000Z", self._bad_time_format_message(source["starting_time"][-7:])
        )
        assert_true(source["ending_time"][-7:] == "000000Z", self._bad_time_format_message(source["ending_time"][-7:]))
        conf = self._configure(
            entity_path=source["entity_path"],
            starting_time=source["starting_time"],
            ending_time=source["ending_time"],
            consumer_group=source.get("consumer_group", "$Default"),
            options=source.get("options"),
        )
        df = self.spark.read.format("eventhubs").options(**conf).load()
        return df

    def read_stream(self, source: dict):
        """Read data from an external input via an Event Hub connection in streaming.

        The source must have the following attributes:

        - "connection": name of the event_hub connection
        - "entity_path": name of the topic to follow
        - "starting_time" or "offset :

            - "starting_time" must use the format '%Y-%m-%dT%H:%M:%S.%fZ'
            - "offset" can be a number wrapped in a string (e.g.:'21235') or '@latest' or '@earliest'
        - "consumer_group" (Optional): event_hub consumer group name
        - "maxEventsPerTrigger" (Optional): Rate limit on maximum number of events processed per trigger interval
        - "options" (Optional): additional options can be added, here is the list of event-hubs-configuration options :
            https://github.com/Azure/azure-event-hubs-spark/blob/master/docs/PySpark/structured-streaming-pyspark.md#event-hubs-configuration

        Because of this bug https://github.com/Azure/azure-event-hubs-spark/issues/444
        we make an assert to prevent retrieving events outside the time bound

        EventHub consumer group documentation:
        https://docs.microsoft.com/en-us/azure/event-hubs/event-hubs-features#event-consumers
        Spark EventHub Connector Documentation:
        https://github.com/Azure/azure-event-hubs-spark/tree/master/docs

        The corresponding event_hub connection must have the following attributes:

        - type = 'event_hub'
        - HubName = name of the hub
        - SharedAccessKeyName = name of the shared access policy (SAS Policy)
        - SharedAccessKey = key of the shared access policy

        :param source: a dictionary describing the source
        :return: a DataFrame (conversationId, destinationAddress, host, message, messageId, messageType, sourceAddress)
        """
        assert_true(source.get("starting_time") or source.get("offset"))
        if "starting_time" in source:
            assert_true(
                source["starting_time"][-7:] == "000000Z", self._bad_time_format_message(source["starting_time"][-7:])
            )
        if "ending_time" in source:
            assert_true(
                source["ending_time"][-7:] == "000000Z", self._bad_time_format_message(source["ending_time"][-7:])
            )
        conf = self._configure(
            entity_path=source["entity_path"],
            starting_time=source.get("starting_time"),
            ending_time=source.get("ending_time"),
            consumer_group=source.get("consumer_group", "$Default"),
            offset=source.get("offset"),
            max_events_per_trigger=source.get("maxEventsPerTrigger"),
            options=source.get("options"),
        )
        df = self.spark.readStream.format("eventhubs").options(**conf).load()
        return df



from karadoc.common.job_core.has_batch_inputs import HasBatchInputs
from karadoc.common.job_core.has_spark import HasSpark
from karadoc.common.job_core.has_vars import HasVars


class AnalyzeJob(HasBatchInputs, HasVars, HasSpark):
    _action_file_name_conf_key = "spark.analyze_timeline"
    _run_method_name = "analyze"

    def __init__(self) -> None:
        HasSpark.__init__(self)
        HasBatchInputs.__init__(self)
        HasVars.__init__(self)

        # Attributes that the user may change
        self.reference_time_col = "application_date"
        self.cohorts = ["cohort"]
        self.nb_buckets = 5
        self.analyze = None



from typing import Generator, Iterable, TypeVar

from karadoc.common.quality.quality_check_job import QualityCheckJob
from karadoc.common.validations import (
    ValidationResult,
    ValidationResultTemplate,
    ValidationSeverity,
)

T = TypeVar("T")


ValidationResult_QualityCheckAlertDuplicateName = ValidationResultTemplate(
    check_type="karadoc.quality_check.alert.duplicate_name",
    message_template='alert name "{alert_name_duplicate}" is defined multiple times',
    default_severity=ValidationSeverity.Error,
)

ValidationResult_QualityCheckMetricDuplicateName = ValidationResultTemplate(
    check_type="karadoc.quality_check.metric.duplicate_name",
    message_template='metric name "{metric_name_duplicate}" is defined multiple times',
    default_severity=ValidationSeverity.Error,
)


def __find_duplicate(_list: Iterable[T]) -> T:
    """Return the first duplicate found in the given list

    >>> __find_duplicate([1, 2, 3, 4, 2, 1])
    2
    >>> __find_duplicate([1, 2, 3, 4])

    >>> __find_duplicate([])


    :param _list:
    :return:
    """
    res = set()
    for elem in _list:
        if elem in res:
            return elem
        else:
            res.add(elem)


def validate_alert_duplicate_names(job: QualityCheckJob, **_) -> Generator[ValidationResult, None, None]:
    """Yield errors if the job has two alerts with the same name"""
    alert_name_duplicate = __find_duplicate([alert.name for alert in job.alerts])
    if alert_name_duplicate is not None:
        yield ValidationResult_QualityCheckAlertDuplicateName(alert_name_duplicate=alert_name_duplicate)


def validate_metric_duplicate_names(job: QualityCheckJob, **_) -> Generator[ValidationResult, None, None]:
    """Yield errors if the job has two metrics with the same name"""
    metric_name_duplicate = __find_duplicate([metric.name for metric in job.metrics])
    if metric_name_duplicate is not None:
        yield ValidationResult_QualityCheckMetricDuplicateName(metric_name_duplicate=metric_name_duplicate)



from typing import Callable, Dict, Optional, Union

from pyspark.sql import DataFrame

from karadoc.common.conf import CONNECTION_GROUP
from karadoc.common.connector import Connector
from karadoc.common.job_core.has_spark import HasSpark


def _read_external_input_signature_check(source: Dict) -> DataFrame:
    """Empty method used to check the signature of the equivalent method defined in the POPULATE files"""
    pass


def _read_external_inputs_signature_check() -> Dict[str, DataFrame]:
    """Empty method used to check the signature of the equivalent method defined in the POPULATE files"""
    pass


class HasExternalInputs(HasSpark):
    def __init__(self) -> None:
        super().__init__()
        # Private attributes
        self.__read_external_input: Optional[Callable] = None
        """Stores the optional override of the read_external_input method by the user"""
        self.__read_external_inputs: Optional[Callable] = None
        """Stores the optional override of the read_external_inputs method by the user"""

        self._limit_external_inputs: Optional[int] = None
        """When set, limits the size of every external input DataFrame to this number of row
        as soon as it is loaded"""

        # Attributes that the user may change in action files
        self.external_inputs: Dict[str, dict] = {}
        """Use this to declare your job's external inputs.
        Expected format is a dictionary of (table alias, table declaration)
        The table description must be a dict specifying the connection to use and all additional parameters specific
        to the external input. The corresponding connection must be properly configured as described in the connector's
        documentation.

        Example:
        {
            "external_input_alias": {
                "{connection}": "name_of_the_connection_used",
                "source_param_name": "source_param_value"
            }
        }
        """

    def _read_external_input_default(self, source: Dict) -> DataFrame:
        connector = self.get_input_connector(source)
        return connector.read(source)

    def _read_external_inputs_default(self) -> Dict[str, DataFrame]:
        return {source_alias: self.read_external_input(source_alias) for source_alias in self.external_inputs}

    def read_external_input(self, source_alias: str) -> DataFrame:
        """Reads a given external input and returns it as a Spark DataFrame

        :param source_alias: the alias of the source in job.external_inputs
        :return: a DataFrame
        """
        source = self.external_inputs[source_alias]
        external_input: DataFrame
        if self.__read_external_input is None:
            external_input = self._read_external_input_default(source)
        else:
            external_input = self.__read_external_input(source)

        if self._limit_external_inputs is not None:
            external_input = external_input.limit(self._limit_external_inputs)
        return external_input

    def read_external_inputs(self) -> Dict[str, DataFrame]:
        """Reads all declared external inputs and returns them as Spark DataFrames

        :return: a Dict[alias, DataFrame]
        """
        external_inputs: Dict[str, DataFrame]
        if self.__read_external_inputs is None:
            external_inputs = self._read_external_inputs_default()
        else:
            external_inputs = self.__read_external_inputs()

        if self._limit_external_inputs is not None:
            for alias, df in external_inputs.items():
                external_inputs[alias] = external_inputs[alias].limit(self._limit_external_inputs)
        return external_inputs

    def load_external_inputs_as_views(self, cache=False):
        for alias in self.external_inputs:
            df = self.read_external_input(alias)
            if cache:
                df = df.cache()
            df.createOrReplaceTempView(alias)

    def get_input_connector(self, source: Union[str, Dict]) -> Connector:
        if type(source) == str:
            source = self.external_inputs[source]
        from karadoc.common.connector import load_connector

        return load_connector(source[CONNECTION_GROUP], self.spark)



import logging
import uuid
from typing import Callable, Dict, Type

from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.exceptions import LoggedException
from karadoc.common.observability import get_logging_context
from karadoc.common.observability.console_event import ConsoleEvent
from karadoc.common.utils.stopwatch import Stopwatch

LOG = logging.getLogger(__name__)


def run_job_with_logging(func: Callable, job_type: Type, model_id: str, job_vars: Dict[str, str], **kwargs):
    logging_context = get_logging_context()
    logging_context.model_id = model_id
    logging_context.job_vars = job_vars
    logging_context.job_type = str(job_type.__name__)
    logging_context.job_uuid = str(uuid.uuid4())
    LOG.user_output(ConsoleEvent(message=f"Starting job for model: {model_id}", event_type="job_start"))
    stopwatch = Stopwatch()
    try:
        func(**{"model_id": model_id, "job_vars": job_vars, **kwargs})
        extra: Dict[str, str] = {"status": "success", "duration_in_seconds": stopwatch.duration_in_seconds}
        LOG.user_output(
            ConsoleEvent(
                message=f"Job ended for model: {model_id}",
                event_type="job_end",
                extra=extra,
                extra_console_message=f"({stopwatch.duration_str})",
            )
        )
        return ReturnCode.Success
    except Exception as err:
        extra: Dict[str, str] = {"status": "failed", "duration_in_seconds": stopwatch.duration_in_seconds}
        LOG.error(
            ConsoleEvent(
                message=f"Job failed for model: {model_id}",
                event_type="job_end",
                extra=extra,
                extra_console_message=f"({stopwatch.duration_str})",
            ),
            exc_info=err,
        )
        raise LoggedException() from err



from argparse import ArgumentParser, Namespace

from karadoc.common.commands.command import Command
from karadoc.common.commands.options.tables_option import TablesOption


class AnalyzeBQCommand(Command):
    description = "compute an excel report with descriptive statistics on bigquery tables"

    @staticmethod
    def add_arguments(parser: ArgumentParser) -> None:
        TablesOption.add_arguments(parser)
        parser.add_argument(
            "--output",
            dest="output",
            type=str,
            default=None,
            metavar="PATH",
            help="Path of the .xlsx file where the results of the command will be written",
        )
        parser.add_argument(
            "--env",
            dest="env",
            type=str,
            default="bigquery",
            metavar="ENV_NAME",
            help="Use credential file from the specified bigquery connection",
        )

    @staticmethod
    def do_command(args: Namespace) -> None:
        # We do this import here to avoid loading unnecessary libraries when running other commands
        from karadoc.biqguery.analyze import analyze_tables

        analyze_tables(args.tables, args.env, args.output)



import sys
import traceback
from typing import TextIO


def display_exception(err: BaseException, file: TextIO = sys.stderr) -> None:
    # This try-catch block is only here to print the error message with real newlines (rather than \n).
    print("Traceback (most recent call last):", file=file)
    traceback.print_tb(err.__traceback__, file=file)
    print(str(err).replace("\\n", "\n"), file=file)

    if err.__cause__:
        print("This exception was caused by another exception: ")
        display_exception(err.__cause__, file)



from typing import Dict, Generator, List, Optional, Tuple

from karadoc.common import conf
from karadoc.common.conf import (
    CONNECTION_GROUP,
    get_conn_conf_id,
    list_connection_names,
)
from karadoc.common.job_core.has_disable import HasDisable
from karadoc.common.job_core.has_external_inputs import HasExternalInputs
from karadoc.common.job_core.has_external_outputs import HasExternalOutputs
from karadoc.common.job_core.has_stream_external_inputs import HasStreamExternalInputs
from karadoc.common.job_core.has_stream_external_output import HasStreamExternalOutput
from karadoc.common.job_core.job_base import JobBase
from karadoc.common.job_core.load import load_non_runnable_action_file
from karadoc.common.model import file_index
from karadoc.common.quality.quality_check_job import QualityCheckJob
from karadoc.common.run.spark_batch_job import SparkBatchJob
from karadoc.common.stream.spark_stream_job import SparkStreamJob
from karadoc.common.validations import (
    ValidationResult,
    ValidationResultTemplate,
    ValidationSeverity,
)

ValidationResult_UnusedConnection = ValidationResultTemplate(
    check_type="karadoc.connection.unused",
    message_template="The connection {conn_conf_id} is not used by any job",
    default_severity=ValidationSeverity.Warning,
)

ValidationResult_ConnectionCouldBeDisabled = ValidationResultTemplate(
    check_type="karadoc.connection.could_be_disabled",
    message_template="The connection {conn_conf_id} is only used by disable jobs and could be disabled",
    default_severity=ValidationSeverity.Warning,
)


def __list_job_io(job: JobBase) -> Generator[dict, None, None]:
    if isinstance(job, HasExternalInputs) or isinstance(job, HasStreamExternalInputs):
        yield from job.external_inputs.values()
    if isinstance(job, HasExternalOutputs):
        yield from job.external_outputs.values()
    if isinstance(job, HasStreamExternalOutput):
        yield job.external_output


def __list_connections_used_by_job(job: JobBase):
    if isinstance(job, HasDisable):
        job_disabled = job.disable
    else:
        job_disabled = False
    for source_details in __list_job_io(job):
        conn = source_details.get(CONNECTION_GROUP)
        if conn is not None:
            yield conn, job_disabled


def __list_used_connections() -> Generator[Tuple[str, bool], None, None]:
    """Return the list of all connections used by action files along with a boolean indicating if the job is disabled"""
    for file_type in [SparkBatchJob, SparkStreamJob, QualityCheckJob]:
        for schema, table, action_file in file_index.list_action_files(file_type):
            full_table_name = schema + "." + table
            job = load_non_runnable_action_file(full_table_name, file_type)
            yield from __list_connections_used_by_job(job)


def _list_used_connections() -> Dict[str, bool]:
    """Return the a dict of all connections used by action files along with a boolean indicating
    if all the job using it are disabled"""
    used_connections = {}
    for conn, job_disabled in __list_used_connections():
        if conn not in used_connections:
            used_connections[conn] = True
        used_connections[conn] &= job_disabled
    return used_connections


def _is_conn_disabled(conn_name, env) -> bool:
    conn_disabled = conf.get_connection_conf(conn_name, env).get("disable")
    if conn_disabled is None:
        conn_disabled = False
    return conn_disabled


def validate_unused_connections(envs: Optional[List[str]]) -> Generator[ValidationResult, None, None]:
    """Return the list of all connections defined in the settings for all the specified envs
    and not used by any action file"""
    used_connections = _list_used_connections()
    if envs is None:
        envs = [None]
    for env in envs:
        for conn_name in list_connection_names(env):
            conn_conf_id = get_conn_conf_id(conn_name, group=CONNECTION_GROUP, env=env)
            if conn_name not in used_connections:
                yield ValidationResult_UnusedConnection(conn_conf_id=conn_conf_id)
            else:
                all_job_using_conn_are_disabled = used_connections[conn_name]
                if not _is_conn_disabled(conn_name, env) and all_job_using_conn_are_disabled:
                    yield ValidationResult_ConnectionCouldBeDisabled(conn_conf_id=conn_conf_id)



import sys
from typing import List

from pyspark.sql import SparkSession

from karadoc.common.commands.exec import run_command
from karadoc.common.commands.return_code import ReturnCode


def main(argv: List[str] = None):
    if argv is None:
        argv = sys.argv[1:]
    return_code = ReturnCode.Error
    try:
        return_code = run_command(argv)
        sys.exit(return_code)
    except Exception:
        sys.exit(return_code)
    finally:
        if SparkSession._instantiatedSession is not None:
            SparkSession._instantiatedSession.stop()



from typing import Generator, Union

from karadoc.common import table_utils
from karadoc.common.job_core.has_inputs import HasInputs
from karadoc.common.job_core.has_keys import HasKeys
from karadoc.common.job_core.has_output import HasOutput
from karadoc.common.job_core.job_base import JobBase
from karadoc.common.validations import (
    ValidationResult,
    ValidationResultTemplate,
    ValidationSeverity,
)

ValidationResult_UnknownInput = ValidationResultTemplate(
    check_type="karadoc.unkown_input",
    message_template="Unknown input table: {table}",
    default_severity=ValidationSeverity.Error,
)

ValidationResult_InvalidOutputPartition = ValidationResultTemplate(
    check_type="karadoc.invalid_output_partition",
    message_template='Output partition "{partition_definition}" is invalid: {cause}',
    default_severity=ValidationSeverity.Error,
)

ValidationResult_SecondaryKeyWithoutPrimaryKey = ValidationResultTemplate(
    check_type="karadoc.secondary_key_without_primary_key",
    message_template="A table should not have secondary keys without a primary key",
    default_severity=ValidationSeverity.Error,
)

ValidationResult_JobLoadError = ValidationResultTemplate(
    check_type="karadoc.job.loading_error",
    message_template="{error_message}:\n{cause}",
    default_severity=ValidationSeverity.Critical,
)


def validate_inputs(job: Union[HasInputs, JobBase], **_) -> Generator[ValidationResult, None, None]:
    """Check that inputs correspond to inputs managed by karadoc"""
    for input_table in job.list_input_tables():
        if not table_utils.table_exists(input_table, type(job)):
            yield ValidationResult_UnknownInput(table=input_table)


def validate_output_partition(job: HasOutput, **_) -> Generator[ValidationResult, None, None]:
    """Check format of output partitions"""
    # TODO #90277 : refactor format of output partitions: dynamic partitions could be dicts
    has_dynamic_partition = False
    for partition_definition in job.output_partition:
        if type(partition_definition) == str:
            if not partition_definition:
                yield ValidationResult_InvalidOutputPartition(
                    partition_definition=partition_definition, cause="empty value"
                )
            has_dynamic_partition = True
        elif type(partition_definition) == tuple:
            if len(partition_definition) != 2:
                yield ValidationResult_InvalidOutputPartition(
                    partition_definition=partition_definition, cause="tuple is not valid (2 elements needed)"
                )
            elif type(partition_definition[0]) != str:
                yield ValidationResult_InvalidOutputPartition(
                    partition_definition=partition_definition, cause="tuple is not valid (first element need to be str)"
                )
            elif not partition_definition[0] or not partition_definition[1]:
                yield ValidationResult_InvalidOutputPartition(
                    partition_definition=partition_definition, cause="empty value"
                )
            elif has_dynamic_partition:
                yield ValidationResult_InvalidOutputPartition(
                    partition_definition=partition_definition,
                    cause="static partition defined after a dynamic partition",
                )
        else:
            yield ValidationResult_InvalidOutputPartition(
                partition_definition=partition_definition, cause="invalid type"
            )


def validate_keys(job: HasKeys, **_) -> Generator[ValidationResult, None, None]:
    """Checks conditions on keys"""
    if len(job.secondary_keys) > 0 and job.primary_key is None:
        yield ValidationResult_SecondaryKeyWithoutPrimaryKey()
